import logging
from typing import Iterable

import torch
from docling_metrics_layout.layout_types import (
    LayoutMetricSample,
    MAPDatasetLayoutEvaluation,
    MAPMetrics,
    MAPPageLayoutEvaluation,
)
from docling_metrics_layout.utils.stats import compute_stats
from docling_metrics_layout.utils.utils import tensor_to_float
from torch import Tensor
from torchmetrics.detection.mean_ap import MeanAveragePrecision

_log = logging.getLogger(__name__)


class MAPLayoutEvaluator:
    def __init__(
        self,
        category_id_to_name: dict[int, str],
    ):
        r""" """
        self._category_id_to_name = category_id_to_name

    def evaluate_sample(
        self,
        sample: LayoutMetricSample,
    ) -> MAPPageLayoutEvaluation:
        r"""
        Evaluation of a single page
        """
        # Extract the targets, predictions from the samples
        targets = []
        predictions = []
        self._extract_from_sample(sample, targets, predictions)

        # Compute mAP
        map_processor = self._get_map_processor()
        map_processor.update(preds=predictions, target=targets)
        map_result = map_processor.compute()

        # Prepare return object with all metrics
        result = MAPPageLayoutEvaluation(
            id=sample.id, **self._export_as_map_metrics(map_result).__dict__
        )
        return result

    def evaluate_dataset(
        self, samples: Iterable[LayoutMetricSample]
    ) -> MAPDatasetLayoutEvaluation:
        r"""Evaluate dataset and compute mAP metrics for all pages"""
        ds_targets = []
        ds_predictions = []
        page_evaluations: dict[str, MAPPageLayoutEvaluation] = {}

        map_values: list[float] = []
        map_50_values: list[float] = []
        map_75_values: list[float] = []

        for sample in samples:
            # Extract targets and predictions for this page
            page_targets = []
            page_predictions = []
            self._extract_from_sample(sample, page_targets, page_predictions)

            # Accumulate for dataset-level metrics
            ds_targets.extend(page_targets)
            ds_predictions.extend(page_predictions)

            # Compute page-level metrics
            map_processor = self._get_map_processor()
            map_processor.update(preds=page_predictions, target=page_targets)
            page_map_result = map_processor.compute()

            page_map_metrics = self._export_as_map_metrics(page_map_result)
            page_evaluation = MAPPageLayoutEvaluation(
                id=sample.id, **page_map_metrics.__dict__
            )
            page_evaluations[sample.id] = page_evaluation

            map_values.append(page_map_metrics.map)
            map_50_values.append(page_map_metrics.map_50)
            map_75_values.append(page_map_metrics.map_75)

        # Compute dataset-level metrics
        map_processor = self._get_map_processor()
        map_processor.update(preds=ds_predictions, target=ds_targets)
        map_result = map_processor.compute()

        ds_evaluation = MAPDatasetLayoutEvaluation(
            page_evaluations=page_evaluations,
            **self._export_as_map_metrics(map_result).__dict__,
            map_stats=compute_stats(map_values),
            map_50_stats=compute_stats(map_50_values),
            map_75_stats=compute_stats(map_75_values),
        )
        return ds_evaluation

    def _extract_from_sample(
        self,
        sample: LayoutMetricSample,
        targets: list,
        predictions: list,
    ):
        r"""
        Extract layout data from the sample and populate the targets/predictions
        """
        # Target
        target_boxes = torch.tensor(
            [bbox.bbox for bbox in sample.page_resolution_a], dtype=torch.float32
        )
        target_labels = torch.tensor(
            [bbox.category_id for bbox in sample.page_resolution_a],
            dtype=torch.int64,
        )
        targets.append(
            {
                "boxes": target_boxes,
                "labels": target_labels,
            }
        )

        # Predictions with confidence scores
        pred_boxes = torch.tensor(
            [bbox.bbox for bbox in sample.page_resolution_b], dtype=torch.float32
        )
        pred_labels = torch.tensor(
            [bbox.category_id for bbox in sample.page_resolution_b],
            dtype=torch.int64,
        )
        pred_scores = torch.tensor(
            [
                bbox.score if bbox.score is not None else 1.0
                for bbox in sample.page_resolution_b
            ],
            dtype=torch.float32,
        )
        predictions.append(
            {
                "boxes": pred_boxes,
                "scores": pred_scores,
                "labels": pred_labels,
            }
        )

    def _export_as_map_metrics(self, map_result: dict) -> MAPMetrics:
        r"""Convert the map_result to MAPMetrics"""
        # Extract scalar metrics
        map_metrics = MAPMetrics(
            map=tensor_to_float(map_result.get("map", -1.0)),
            map_50=tensor_to_float(map_result.get("map_50", -1.0)),
            map_75=tensor_to_float(map_result.get("map_75", -1.0)),
            map_large=tensor_to_float(map_result.get("map_large", -1.0)),
            map_medium=tensor_to_float(map_result.get("map_medium", -1.0)),
            map_small=tensor_to_float(map_result.get("map_small", -1.0)),
            mar_1=tensor_to_float(map_result.get("mar_1", -1.0)),
            mar_10=tensor_to_float(map_result.get("mar_10", -1.0)),
            mar_100=tensor_to_float(map_result.get("mar_100", -1.0)),
            mar_large=tensor_to_float(map_result.get("mar_large", -1.0)),
            mar_medium=tensor_to_float(map_result.get("mar_medium", -1.0)),
            mar_small=tensor_to_float(map_result.get("mar_small", -1.0)),
            # Extract per-class metrics
            map_per_class=self._extract_metrics_per_class(
                map_result.get("classes", None), map_result.get("map_per_class", None)
            ),
            mar_100_per_class=self._extract_metrics_per_class(
                map_result.get("classes", None),
                map_result.get("mar_100_per_class", None),
            ),
        )
        return map_metrics

    def _extract_metrics_per_class(
        self, classes_tensor: Tensor, per_class_tensor: Tensor
    ) -> dict[str, float]:
        r"""
        Extract per-class metrics from tensor and map to class names
        """
        if per_class_tensor is None:
            return {}

        evaluated_classes: list[int] = (
            [classes_tensor.tolist()]
            if classes_tensor.numel() == 1
            else classes_tensor.tolist()
        )
        map_classes = (
            [per_class_tensor.tolist()]
            if per_class_tensor.numel() == 1
            else per_class_tensor.tolist()
        )
        per_class_dict: dict[str, float] = {}
        for idx, category_map in enumerate(map_classes):
            category_id = evaluated_classes[idx]
            category_name = self._category_id_to_name[category_id]
            per_class_dict[category_name] = category_map
        return per_class_dict

    def _get_map_processor(self) -> MeanAveragePrecision:
        r"""Factory of  MeanAveragePrecision"""
        return MeanAveragePrecision(
            box_format="xyxy",
            iou_type="bbox",
            class_metrics=True,
            backend="faster_coco_eval",
        )



import logging
from concurrent.futures import Future, ProcessPoolExecutor, as_completed
from pathlib import Path
from typing import Iterable, Optional

import numpy as np
from tqdm import tqdm  # type: ignore

from docling_metrics_layout.layout_types import (
    BboxResolution,
    DatasetToreLayoutEvaluation,
    LayoutMetricSample,
    MultiLabelMatrixEvaluation,
    PageToreEvaluation,
)
from docling_metrics_layout.tore.confusion_matrix_exporter import (
    ConfusionMatrixExporter,
)
from docling_metrics_layout.tore.multi_label_confusion_matrix import (
    MultiLabelConfusionMatrix,
)

_log = logging.getLogger(__name__)


def evaluate_page(
    mlcm: MultiLabelConfusionMatrix,
    id: str,
    pg_width: int,
    pg_height: int,
    matrix_id_to_name: dict[int, str],
    page_resolutions_a: list[BboxResolution],
    page_resolutions_b: Optional[list[BboxResolution]] = None,
) -> tuple[str, int, MultiLabelMatrixEvaluation]:
    r"""
    Compute the confusion matrix and the metrics for one page
    If pred_resolutions is None, assume an all-background predictions

    Return
    ------
    page_pixels
    page_metrics
    """
    # Make binary representations
    gt_binary = mlcm.make_binary_representation(pg_width, pg_height, page_resolutions_a)
    if page_resolutions_b is not None:
        preds_binary = mlcm.make_binary_representation(
            pg_width, pg_height, page_resolutions_b
        )
    else:
        preds_binary = np.ones((pg_height, pg_width), dtype=np.uint64)

    # Compute confusion matrix
    matrix_categories_ids: list[int] = list(matrix_id_to_name.keys())
    confusion_matrix = mlcm.generate_confusion_matrix(
        gt_binary, preds_binary, matrix_categories_ids
    )

    # Compute metrics
    page_metrics: MultiLabelMatrixEvaluation = mlcm.compute_metrics(
        confusion_matrix, matrix_id_to_name
    )
    page_pixels = pg_width * pg_height

    return id, page_pixels, page_metrics


class ToreLayoutEvaluator:
    r"""
    TORE: Taxonomy-invariant Object Recognition Evaluation
    """

    def __init__(
        self,
        category_id_to_name: dict[int, str],
        concurrency: int,
    ):
        r"""
        Parameters:
        -----------
        category_id_to_name: Mapping of category ids to category names
                             This mapping must NOT include any Background label
        concurrency: Parallelism used when the
        """
        self._category_id_to_name = category_id_to_name
        self._concurrency = concurrency

        # Initialize the multi label confusion matrix calculator
        self._mlcm = MultiLabelConfusionMatrix(validation_mode="disabled")

        # Build matrix categories with background at index 0
        self._matrix_id_to_name: dict[int, str]  # Matrix ID to category name

        self._category_id_to_matrix_id: dict[
            int, int
        ]  # Original category ID to matrix ID

        self._matrix_id_to_category_id: dict[
            int, int
        ]  # Matrix ID to original category ID

        (
            self._matrix_id_to_name,
            self._category_id_to_matrix_id,
            self._matrix_id_to_category_id,
        ) = self._build_matrix_categories()

    @staticmethod
    def evaluation_filenames(save_root: Path) -> dict[str, Path]:
        r"""
        Generate the expected filenames for the produced evaluation files
        """
        json_fn = save_root / "evaluation_pixel_layout.json"
        excel_fn = save_root / "evaluation_pixel_layout.xlsx"

        eval_filenames: dict[str, Path] = {
            "json": json_fn,
            "excel": excel_fn,
        }
        return eval_filenames

    def _build_matrix_categories(
        self,
    ) -> tuple[dict[int, str], dict[int, int], dict[int, int]]:
        r"""
        Returns:
        --------
        matrix_id_to_name: dict[int, str]
        category_id_to_matrix_id: dict[int, int]
        matrix_id_to_category_id: dict[int, int]
        """
        matrix_id_to_name: dict[int, str] = {0: "Background"}
        category_id_to_matrix_id: dict[int, int] = {}
        matrix_id_to_category_id: dict[int, int] = {}

        for matrix_id, (category_id, name) in enumerate(
            self._category_id_to_name.items(), start=1
        ):
            matrix_id_to_name[matrix_id] = name
            category_id_to_matrix_id[category_id] = matrix_id
            matrix_id_to_category_id[matrix_id] = category_id

        return matrix_id_to_name, category_id_to_matrix_id, matrix_id_to_category_id

    def evaluate_sample(
        self,
        sample: LayoutMetricSample,
    ) -> PageToreEvaluation:
        r"""
        Evaluation of a single page
        """
        page_pixels: int
        page_metrics: MultiLabelMatrixEvaluation
        _, page_pixels, page_metrics = evaluate_page(
            self._mlcm,
            sample.id,
            sample.page_width,
            sample.page_height,
            self._matrix_id_to_name,
            sample.page_resolution_a,
            sample.page_resolution_b,
        )
        return PageToreEvaluation(
            id=sample.id,
            num_pixels=page_pixels,
            matrix_evaluation=page_metrics,
        )

    def evaluate_dataset(
        self, samples: Iterable[LayoutMetricSample]
    ) -> DatasetToreLayoutEvaluation:
        r""" """
        matrix_categories_ids: list[int] = list(self._matrix_id_to_name.keys())
        num_categories = len(matrix_categories_ids)
        ds_confusion_matrix = np.zeros((num_categories, num_categories))
        all_pages_evaluations: dict[
            str, PageToreEvaluation
        ] = {}  # Key is doc_id-page-no
        ds_num_pixels = 0
        pages_detailed_f1: list[
            float
        ] = []  # Gather f1 score/image when evaluated on all classes
        pages_collapsed_f1: list[
            float
        ] = []  # Gather f1 score/image when evaluated on collapsed classes

        evaluated_samples = 0
        with ProcessPoolExecutor(max_workers=self._concurrency) as executor:
            futures: list[Future] = []

            # Submit pages for execution
            _log.info("Submitting the documents for evaluation...")
            for sample in samples:
                evaluated_samples += 1

                # Submit the page for computation
                futures.append(
                    executor.submit(
                        evaluate_page,
                        self._mlcm,
                        sample.id,
                        sample.page_width,
                        sample.page_height,
                        self._matrix_id_to_name,
                        sample.page_resolution_a,
                        sample.page_resolution_b,
                    )
                )

            # Collect the futures
            _log.info("Collecting the pages for evaluations...")
            for future in tqdm(
                as_completed(futures),
                desc="Multi-label Matrix Layout evaluations",
                ncols=120,
                total=len(futures),
            ):
                page_metrics: MultiLabelMatrixEvaluation
                doc_page_id, page_pixels, page_metrics = future.result()

                page_confusion_matrix: np.ndarray = (
                    page_metrics.detailed.confusion_matrix
                )
                ds_num_pixels += page_pixels
                ds_confusion_matrix += page_confusion_matrix
                page_evaluation = PageToreEvaluation(
                    id=doc_page_id,
                    num_pixels=page_pixels,
                    matrix_evaluation=page_metrics,
                )
                all_pages_evaluations[doc_page_id] = page_evaluation

                # Update f1 lists
                pages_detailed_f1.append(
                    page_metrics.detailed.agg_metrics.classes_f1_mean
                )
                pages_collapsed_f1.append(
                    page_metrics.collapsed.agg_metrics.classes_f1_mean
                )

        # Compute metrics for the dataset
        ds_matrix_evaluation: MultiLabelMatrixEvaluation = self._mlcm.compute_metrics(
            ds_confusion_matrix,
            self._matrix_id_to_name,
        )

        ds_evaluation = DatasetToreLayoutEvaluation(
            evaluated_samples=evaluated_samples,
            num_pages=len(all_pages_evaluations),
            num_pixels=ds_num_pixels,
            matrix_evaluation=ds_matrix_evaluation,
            page_evaluations=all_pages_evaluations,
        )

        return ds_evaluation

    def export_evaluations(
        self,
        ds_evaluation: DatasetToreLayoutEvaluation,
        save_root: Path,
        export_excel_reports: bool = True,
    ):
        r"""
        Save all evaluations as jsons and excel reports
        """
        save_root.mkdir(parents=True, exist_ok=True)

        # Get the evaluation filenames
        eval_fns = ToreLayoutEvaluator.evaluation_filenames(save_root)

        # # Save the dataset evaluation as a json
        # json_fn = eval_fns["json"]
        # with open(json_fn, "w") as fd:
        #     json.dump(ds_evaluation.model_dump(), fd, indent=2, sort_keys=True)

        # Export excel reports
        if not export_excel_reports:
            return

        excel_exporter = ConfusionMatrixExporter()
        headers = list(self._matrix_id_to_name.values())
        collapsed_headers: list[str] = [
            f"{metric}: {cell}"
            for metric in ["Precision(GT/Pred)", "Recall(GT/Pred)", "F1(GT/Pred)"]
            for cell in [
                "BG/BG",
                "BG/cls",
                "cls/BG",
                "cls/cls",
            ]
        ]
        image_collapsed_aggs: dict[str, np.ndarray] = {}
        for doc_page_id, page_evaluations in ds_evaluation.page_evaluations.items():
            pm = page_evaluations.matrix_evaluation.collapsed
            if not pm:
                continue
            # [12,]
            image_collapsed_vector = np.stack(
                [
                    pm.precision_matrix.flatten(),
                    pm.recall_matrix.flatten(),
                    pm.f1_matrix.flatten(),
                ],
                axis=0,
            ).flatten()
            image_collapsed_aggs[doc_page_id] = image_collapsed_vector

        excel_fn = eval_fns["excel"]

        title = "Pixel-wise Multi-Label Confusion Matrix Evaluations"
        excel_exporter.build_ds_report(
            title,
            ds_evaluation.num_pages,
            ds_evaluation.num_pixels,
            headers,
            ds_evaluation.matrix_evaluation,
            collapsed_headers,
            image_collapsed_aggs,
            excel_fn,
        )



import argparse
import json
import logging
import time
from datetime import datetime
from pathlib import Path
from statistics import mean, median
from typing import Any

from docling_metrics_layout import LayoutMetrics
from docling_metrics_layout.benchmarks.tools import load_layout_samples
from docling_metrics_layout.layout_types import LayoutMetricSample

_log = logging.getLogger(__name__)


class Benchmarker:
    r"""
    Benchmark layout evaluation runtime on a COCO-format dataset.

    Measures the runtime of each private _evaluate_xxx method of LayoutMetrics:
      - _evaluate_tore_sample  (per-sample)
      - _evaluate_map_sample    (per-sample)
      - _evaluate_tore_dataset (dataset-level)
      - _evaluate_map_dataset   (dataset-level)
    """

    def __init__(self, save_root: Path, concurrency: int = 4) -> None:
        r""" """
        self._save_root = save_root
        self._save_root.mkdir(parents=True, exist_ok=True)
        self._concurrency = concurrency

    def benchmark(
        self,
        gt_coco_fn: Path,
        preds_coco_fn: Path,
        limit: int | None = None,
    ):
        r""" """
        _log.info("Convert COCO data to internal representation...")
        category_id_to_name, samples = load_layout_samples(gt_coco_fn, preds_coco_fn)
        if limit is not None:
            samples = samples[:limit]
        _log.info("Samples to evaluate: %d", len(samples))

        lm = LayoutMetrics(
            category_id_to_name=category_id_to_name,
            concurrency=self._concurrency,
        )

        report: dict[str, Any] = {
            "concurrency": self._concurrency,
            "stats": {},
            "samples": {},
            "dataset": {},
        }

        if False:
            self._evaluate_individual_samples(samples, lm, report)

        # Measure dataset-level TORE evaluation time
        _log.info("Benchmarking TORE metrics for the entire dataset...")
        t0 = time.perf_counter()
        tore_dataset_eval = lm._evaluate_tore_dataset(samples)  # noqa: F841
        tore_dataset_ms = (time.perf_counter() - t0) * 1000
        n = len(samples)
        report["dataset"]["size"] = n
        report["dataset"]["tore_dataset"] = {
            "ms": tore_dataset_ms,
            "average_ms": tore_dataset_ms / n if n else 0.0,
            # Debug: Disable dumping the full metrics
            # "metrics": tore_dataset_eval.model_dump(),
        }
        _log.info(
            "tore_dataset: %.2fms for %d samples (%.4fms/sample)",
            tore_dataset_ms,
            n,
            tore_dataset_ms / n if n else 0.0,
        )

        # Measure dataset-level mAP evaluation time
        _log.info("Benchmarking mAP metrics for the entire dataset...")
        t0 = time.perf_counter()
        map_dataset_eval = lm._evaluate_map_dataset(samples)  # noqa: F841
        map_dataset_ms = (time.perf_counter() - t0) * 1000
        report["dataset"]["map_dataset"] = {
            "ms": map_dataset_ms,
            "average_ms": map_dataset_ms / n if n else 0.0,
            # Debug: Disable dumping the full metrics
            # "metrics": map_dataset_eval.model_dump(),
        }
        _log.info(
            "map_dataset: %.2fms for %d samples (%.4fms/sample)",
            map_dataset_ms,
            n,
            map_dataset_ms / n if n else 0.0,
        )

        # Save report
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        report_path = self._save_root / f"benchmark_report_{timestamp}.json"
        with open(report_path, "w") as f:
            json.dump(report, f, indent=2, sort_keys=True)
        _log.info("Benchmark report saved to: %s", report_path)

    def _evaluate_individual_samples(
        self,
        samples: list[LayoutMetricSample],
        lm: LayoutMetrics,
        report: dict[str, Any],
    ):
        r""" """
        _log.info("Benchmarking metrics for individual samples...")

        # Collect per-sample timing data
        timing_data: dict[str, list[float]] = {
            "tore_sample": [],
            "map_sample": [],
        }
        for sample in samples:
            id = sample.id
            sample_bench: dict[str, dict[str, float]] = {}

            # Measure TORE sample evaluation time
            t0 = time.perf_counter()
            lm._evaluate_tore_sample(sample)
            tore_sample_ms = (time.perf_counter() - t0) * 1000
            sample_bench["tore_sample"] = {"ms": tore_sample_ms}
            timing_data["tore_sample"].append(tore_sample_ms)

            # Measure mAP sample evaluation time
            t0 = time.perf_counter()
            lm._evaluate_map_sample(sample)
            map_sample_ms = (time.perf_counter() - t0) * 1000
            sample_bench["map_sample"] = {"ms": map_sample_ms}
            timing_data["map_sample"].append(map_sample_ms)

            report["samples"][id] = sample_bench
            _log.info(
                "%s | tore_sample: %.2fms | map_sample: %.2fms",
                id,
                tore_sample_ms,
                map_sample_ms,
            )

        # Compute statistics for per-sample methods
        for task_name, times in timing_data.items():
            if times:
                report["stats"][task_name] = {
                    "min": min(times),
                    "max": max(times),
                    "mean": mean(times),
                    "median": median(times),
                }
                _log.info(
                    "%s stats | min: %.2fms | max: %.2fms | mean: %.2fms | median: %.2fms",
                    task_name,
                    min(times),
                    max(times),
                    mean(times),
                    median(times),
                )
        return report


def main():
    parser = argparse.ArgumentParser(description="Layout Metrics Benchmark")
    parser.add_argument(
        "-G",
        "--gt_coco",
        type=Path,
        required=True,
        help="Path to the COCO ground-truth annotations JSON file",
    )
    parser.add_argument(
        "-p",
        "--preds_coco",
        type=Path,
        required=True,
        help="Path to the COCO predictions JSON file",
    )
    parser.add_argument(
        "-s",
        "--save_root",
        type=Path,
        required=True,
        help="Path to the directory to save generated data",
    )
    parser.add_argument(
        "-l",
        "--limit",
        type=int,
        default=None,
        help="Maximum number of samples to benchmark (default: all)",
    )
    parser.add_argument(
        "-c",
        "--concurrency",
        type=int,
        default=4,
        help="Number of concurrent workers for TORE (default: 4)",
    )
    args = parser.parse_args()

    log_format = "%(asctime)s - %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=log_format)

    _log.info("Benchmark LayoutMetrics implementations")
    benchmarker = Benchmarker(args.save_root, concurrency=args.concurrency)
    benchmarker.benchmark(
        gt_coco_fn=args.gt_coco,
        preds_coco_fn=args.preds_coco,
        limit=args.limit,
    )


if __name__ == "__main__":
    main()



import logging
from enum import Enum
from pathlib import Path
from typing import Iterable, Optional

from docling_metrics_core.base_types import (
    BaseMetric,
)

from docling_metrics_layout.layout_types import (
    DatasetToreLayoutEvaluation,
    LayoutMetricDatasetEvaluation,
    LayoutMetricSample,
    LayoutMetricSampleEvaluation,
    MAPDatasetLayoutEvaluation,
    MAPPageLayoutEvaluation,
    PageToreEvaluation,
)
from docling_metrics_layout.map.map_layout_evaluator import (
    MAPLayoutEvaluator,
)
from docling_metrics_layout.tore.tore_layout_evaluator import (
    ToreLayoutEvaluator,
)

_log = logging.getLogger(__name__)

# Silence the coco tools
logging.getLogger("faster_coco_eval").setLevel(logging.WARNING)


class LayoutMetricsMode(str, Enum):
    PYTHON = "Python"
    CPP = "C++"


class LayoutMetrics(BaseMetric):
    r"""
    Various text metrics
    """

    def __init__(
        self,
        category_id_to_name: dict[int, str],
        concurrency: int = 4,
        save_root: Optional[Path] = None,
        mode: LayoutMetricsMode = LayoutMetricsMode.PYTHON,
    ):
        r"""
        Initialize the LayoutMetrics evaluator.

        Parameters:
            category_id_to_name: Mapping of category IDs to their string names.
            concurrency: Number of concurrent workers for parallel evaluation (default: 4).
            save_root: Optional root directory path for saving evaluation results.
                       If None, results will not be saved to disk.
            mode: Execution mode for layout evaluation, either PYTHON or C++ (default: PYTHON).
        """
        self._save_root = save_root
        self._category_id_to_name = category_id_to_name
        self._mode = mode

        # Evaluators
        self._tore_evaluator = ToreLayoutEvaluator(category_id_to_name, concurrency)
        self._map_evaluator = MAPLayoutEvaluator(category_id_to_name)

    def evaluate_sample(
        self, sample: LayoutMetricSample
    ) -> LayoutMetricSampleEvaluation:
        r"""Evaluate a single sample with TORE and mAP metrics"""
        # Evaluate TORE metric
        page_tore_evaluation = self._evaluate_tore_sample(sample)

        # Evaluate mAP metric
        page_map_layout_evaluation = self._evaluate_map_sample(sample)

        sample_evaluation = LayoutMetricSampleEvaluation(
            id=sample.id,
            page_tore_evaluation=page_tore_evaluation,
            page_map_layout_evaluation=page_map_layout_evaluation,
        )
        return sample_evaluation

    def aggregate(
        self, results: Iterable[LayoutMetricSampleEvaluation]
    ) -> Optional[LayoutMetricDatasetEvaluation]:
        r""" """
        return None

    def evaluate_dataset(
        self, samples: Iterable[LayoutMetricSample]
    ) -> LayoutMetricDatasetEvaluation:
        r"""Evaluate a dataset with TORE and mAP metrics"""
        sample_list = list(samples)

        # Evaluate mAP metrics
        ds_map_layout_evaluation = self._evaluate_map_dataset(sample_list)

        # Evaluate TORE metrics
        ds_tore_evaluation = self._evaluate_tore_dataset(sample_list)

        # Save export
        reports: list[Path] = []
        if self._save_root:
            _log.info("Exporting TORE evalution in %s", str(self._save_root))
            self._tore_evaluator.export_evaluations(ds_tore_evaluation, self._save_root)
            reports.append(
                ToreLayoutEvaluator.evaluation_filenames(self._save_root)["excel"]
            )

        # Build return object
        result = LayoutMetricDatasetEvaluation(
            sample_count=len(sample_list),
            dataset_tore_evaluation=ds_tore_evaluation,
            dataset_map_layout_evaluation=ds_map_layout_evaluation,
            reports=reports,
        )

        return result

    def _evaluate_tore_sample(self, sample: LayoutMetricSample) -> PageToreEvaluation:
        r"""Evaluate TORE metrics for a single sample"""
        _log.info("Evaluate TORE metrics for a single sample")
        return self._tore_evaluator.evaluate_sample(sample)

    def _evaluate_map_sample(
        self, sample: LayoutMetricSample
    ) -> MAPPageLayoutEvaluation:
        r"""Evaluate mAP metrics for a single sample"""
        _log.info("Evaluate mAP layout metrics for a single sample")
        return self._map_evaluator.evaluate_sample(sample)

    def _evaluate_tore_dataset(
        self, samples: list[LayoutMetricSample]
    ) -> DatasetToreLayoutEvaluation:
        r"""Evaluate TORE for a dataset"""
        _log.info("Evaluate TORE metrics for a dataset")
        return self._tore_evaluator.evaluate_dataset(samples)

    def _evaluate_map_dataset(
        self, samples: list[LayoutMetricSample]
    ) -> MAPDatasetLayoutEvaluation:
        r"""Evaluate mAP metrics for a dataset"""
        _log.info("Evaluate mAP layout metrics for a dataset")
        return self._map_evaluator.evaluate_dataset(samples)



from torch import Tensor


def dict_get(data: dict, keys: list[str], default=None):
    r"""
    Traverse the given path of keys and return the value of dict
    If the path is broken return the default value
    """
    current = data
    for key in keys:
        if isinstance(current, dict) and key in current:
            current = current[key]
        else:
            return default
    return current


def xyxy_to_xywh(bbox: list[float]) -> list[float]:
    r"""
    The reverse of wywh_to_xyxy

    Assumptions:
    - Both input and output bbox is defined with the image origin at the top-left corner.
    - The input bbox is [x1, y1, x2, y2] where [x1, y1] is the top-left corner and [x2, y2] the
      bottom-right corner.
    - The output bbox is [x1, y1, width, height]
    """
    width = bbox[2] - bbox[0]
    height = bbox[3] - bbox[1]
    xywh_bbox = [bbox[0], bbox[1], width, height]
    return xywh_bbox


def xywh_to_xyxy(bbox: list[float]) -> list[float]:
    r"""
    The reverse of xyxy_to_xywh

    Assumptions:
    - Both input and output bbox is defined with the image origin at the top-left corner.
    - The input bbox is [x1, y1, x2, y2] where [x1, y1] is the top-left corner and [x2, y2] the bottom-right corner.
    - The output bbox is [x1, y1, w, h]
    """
    xyxy_bbox = [
        bbox[0],
        bbox[1],
        bbox[0] + bbox[2],
        bbox[1] + bbox[3],
    ]
    return xyxy_bbox


def tensor_to_float(t: Tensor | float) -> float:
    r"""Get float from tensor item"""
    if isinstance(t, Tensor):
        return float(t.item())
    return t



import logging
import math
from typing import Any, Optional

import numpy as np

from docling_metrics_layout.layout_types import (
    BboxResolution,
    MultiLabelMatrixAggMetrics,
    MultiLabelMatrixEvaluation,
    MultiLabelMatrixMetrics,
)

_log = logging.getLogger(__name__)


def unpackbits(x: np.ndarray, num_bits: int):
    r"""
    Unpack num_bits bits of each element of the numpy array x
    The number of bits defines how many bits we will take from x to unpack.
    """
    xshape = list(x.shape)
    x = x.reshape([-1, 1])
    mask: np.ndarray = 2 ** np.arange(num_bits, dtype=x.dtype).reshape([1, num_bits])
    return (x & mask).astype(bool).astype(int).reshape(xshape + [num_bits])


def compress_binary_representations(
    gt: np.ndarray,
    preds: np.ndarray,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    r"""
    1. Combine element-wise pairs from gt and preds (G, P)
    2. Find the unique pairs and the counts c of each pair.
    3. Decouple the unique pairs into 2 flattened arrays g and p.
    4. Return the g, p, c

    Parameters:
    -----------
    gt: [num_cat, num_cat]
    preds: [num_cat, num_cat]

    Returns:
    --------
    g: [u,]: The flattened values from gt that come from a unique pair (g, p)
    p: [u,]: The flattened values from preds that come from a unique pair (g, p)
    c: [u,]: The counts for each unique pair (g, p)
    """
    # Build a structured array to keep the pairs (g, p)
    pairs = np.zeros(gt.shape, dtype=[("gt", gt.dtype), ("preds", preds.dtype)])
    pairs["gt"] = gt
    pairs["preds"] = preds
    u, counts = np.unique(pairs, return_counts=True)
    g = u["gt"]
    p = u["preds"]
    return g, p, counts


class MultiLabelConfusionMatrix:
    r""" """

    DETAILED_METRICS_KEY = "detailed_classes"
    COLLAPSED_METRICS_KEY = "collapsed_classes"
    ALL_COLLAPSED_CLASSES_NAME = "all_classes"

    def __init__(
        self,
        validation_mode: str = "disabled",
    ):
        r"""
        The validation mode can be one of: ["disabled", "log", "raise"]
        """
        self._validation_mode = validation_mode

    def make_binary_representation(
        self,
        image_width: int,
        image_height: int,
        resolutions: list[BboxResolution],
        set_background: bool = True,
    ) -> np.ndarray:
        r"""
        Create a numpy matrix with the binary representation of the layout resolutions
        Each pixel is represented as one uint64 integer, where the 1-bit flags presence of a class.

        Parameters
        ----------
        set_background: Assign the value 1 to all pixels that still have a zero value in the end

        Returns
        -------
        np.ndarray with the binary representation of the resolutions. Dims are equal to the image size
        """
        # Initialize the representation matrix with 0
        matrix: np.ndarray = np.zeros(
            (
                image_height,
                image_width,
            ),
            dtype=np.uint64,
        )

        for res in resolutions:
            x1 = res.bbox[0]
            y1 = res.bbox[1]
            x2 = res.bbox[2]
            y2 = res.bbox[3]
            x_begin = math.floor(x1)
            x_end = math.ceil(x2)
            y_begin = math.floor(y1)
            y_end = math.ceil(y2)

            cat_id = res.category_id
            bit_index = np.uint64(1 << cat_id)
            matrix[y_begin:y_end, x_begin:x_end] |= bit_index

        # Set the background class (binary 1) if there is no other class set
        if set_background:
            matrix[matrix == 0] = 1

        return matrix

    def generate_confusion_matrix(
        self,
        gt: np.ndarray,
        preds: np.ndarray,
        categories: list[int],
    ) -> np.ndarray:
        r"""
        Create the confusion matrix for multi-label predictions.
        The returned matrix can be used to compute precision, recall.

        In a perfect prediction the matrix should be diagonal. All elements outside of the main
        diagonal indicate prediction errors and contribute to penalties in the calculation of
        precision, recall.

        Inspired by "Multi-label classifier performance evaluation with confusion matrix"
        https://csitcp.org/paper/10/108csit01.pdf

        Parameters:
        -----------
        gt: GT binary data representation. Each value is the bit-encoding of the classes per pixel
        preds: Preds binary data representation. Each value is the bit-encoding of the classes per pixel
        categories: list[category_id]

        Returns
        -------
        np.ndarray [num_categories + 1, num_categories + 1]. The +1 is for the background class
        """
        # Compress the binary representations pair-wise
        comp_gt, comp_preds, counts = compress_binary_representations(gt, preds)
        _log.debug("Original dims: %s, compressed dims: %s", gt.shape, comp_gt.shape)

        # Compute the confusion matrix
        confusion_matrix = self._compute_confusion_matrix(
            comp_gt,
            comp_preds,
            categories,
            weights=counts,
        )
        return confusion_matrix

    def _compute_confusion_matrix(
        self,
        gt: np.ndarray,
        preds: np.ndarray,
        categories: list[int],
        weights: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        r"""
        gt, preds, weights must have exactly the same dimensions, no matter what that dimension is
        """
        num_categories = len(categories)

        # confusion_matrix: [num_categories, num_categories]
        confusion_matrix: np.ndarray = np.zeros(
            (num_categories, num_categories), dtype=float
        )
        eye = np.eye(num_categories)

        ############################################################################################
        # Case 1: Perfect prediction
        #

        # [img_height, img_width]
        selections_case1 = gt == preds
        case1_gt_pixels = gt[selections_case1]

        # 1. I = np.eye(num_categories): [num_categories, num_categories]
        # 2. U = unpackbits(gt[selections_case1], num_categories)]: [k, num_categories]
        # 3. C = U[:, None, :] * I[None, :, :]
        # U = unpackbits(gt[selections_case1], num_categories)
        # C = U[:, None, :] * eye[None, :, :]

        # case1_contributions: [num_pixels_with_perfect_preds, num_categories, num_categories]
        case1_contributions = (
            unpackbits(case1_gt_pixels, num_categories)[:, None, :] * eye[None, :, :]
        )
        # print("Case1 contributions:")
        # print(case1_contributions)

        # Validate the contributions
        self._validate_contributions(case1_gt_pixels, case1_contributions, "Case1")

        if weights is not None:
            case1_weights = weights[selections_case1]
            case1_contributions = case1_contributions * case1_weights[:, None, None]
        confusion_matrix += np.sum(case1_contributions, axis=0)

        ############################################################################################
        # Case 2: Prediction has all GT plus extra mistakes
        #

        # Filter out the non-perfect predictions to take the ones where preds contain all GT bits
        # [img_height, img_width]
        selections_case2 = ~selections_case1
        # selections_case2[selections_case2 == True] = (
        selections_case2[selections_case2] = (
            gt[selections_case2] & preds[selections_case2] == gt[selections_case2]
        )

        # [num_pixels_with_extra_preds,]
        case2_preds_pixels = preds[selections_case2]
        case2_gt_pixels = gt[selections_case2]

        # [num_pixels_with_extra_preds, num_categories]
        case2_preds_gt_intersection = case2_preds_pixels & case2_gt_pixels
        case2_preds_gt_intersection = unpackbits(
            case2_preds_gt_intersection, num_categories
        )

        if len(case2_preds_pixels) > 0:
            # [num_pixels_with_extra_preds, num_categories]
            case2_preds_gt_diff = (
                case2_preds_pixels ^ case2_gt_pixels
            ) & case2_preds_pixels
            case2_preds_gt_diff = unpackbits(case2_preds_gt_diff, num_categories)

            # [num_pixels_with_extra_preds, num_categories, num_categories]
            case2_penalty = (
                case2_preds_gt_intersection[:, :, None]
                * case2_preds_gt_diff[:, None, :]
            )

            # [num_pixels_with_extra_preds, num_categories, num_categories]
            case2_gt_diagonals = (
                unpackbits(case2_gt_pixels, num_categories)[:, None, :]
                * eye[None, :, :]
            )

            # [num_pixels_with_extra_preds,]
            case2_gt_multiplier = np.bitwise_count(case2_gt_pixels)

            # [num_pixels_with_extra_preds,]
            case2_preds_divider = np.bitwise_count(case2_preds_pixels)

            # [num_pixels_with_extra_preds, num_categories, num_categories]
            case2_gain = case2_gt_multiplier[:, None, None] * case2_gt_diagonals

            # [num_pixels_with_extra_preds, num_categories, num_categories]
            case2_contributions = (case2_penalty + case2_gain) / case2_preds_divider[
                :, None, None
            ]
            # print("Case2 contributions:")
            # print(case2_contributions)

            # Validate the contributions
            self._validate_contributions(case2_gt_pixels, case2_contributions, "Case2")

            if weights is not None:
                case2_weights = weights[selections_case2]
                case2_contributions = case2_contributions * case2_weights[:, None, None]
            confusion_matrix += np.sum(case2_contributions, axis=0)

        ############################################################################################
        # Case 3: GT has more labels than preds
        # NOTICE: This case NEVER happens for us because our GT has only 1 label
        #

        # [img_height, img_width]
        selections_case3 = ~selections_case1
        # selections_case3[selections_case3 == True] = (
        selections_case3[selections_case3] = (
            gt[selections_case3] | preds[selections_case3] == gt[selections_case3]
        )

        # [num_pixels_with_additional_gt_labels,]
        case3_preds_pixels = preds[selections_case3]
        if len(case3_preds_pixels) > 0:
            case3_gt_pixels = gt[selections_case3]

            # [num_pixels_with_additional_gt_labels, num_categories]
            case3_gt_preds_diff = (case3_preds_pixels ^ case3_gt_pixels) & gt[
                selections_case3
            ]
            case3_gt_preds_diff = unpackbits(case3_gt_preds_diff, num_categories)

            # [num_pixels_with_additional_gt_labels,]
            case3_preds_divider = np.bitwise_count(case3_preds_pixels)

            # [num_pixels_with_additional_gt_labels, num_categories, num_categories]
            case3_preds_diagonals = (
                unpackbits(case3_preds_pixels, num_categories)[:, None, :]
                * eye[None, :, :]
            )

            # [num_pixels_with_additional_gt_labels, num_categories]
            case3_preds = unpackbits(case3_preds_pixels, num_categories)

            # [num_pixels_with_additional_gt_labels, num_categories, num_categories]
            case3_penalty = (
                case3_gt_preds_diff[:, :, None] * case3_preds[:, None, :]
            ) / case3_preds_divider[:, None, None]

            # [num_pixels_with_additional_gt_labels, num_categories, num_categories]
            case3_contributions = case3_penalty + case3_preds_diagonals
            # print("Case3 contributions:")
            # print(case3_contributions)

            # Validate the contributions
            self._validate_contributions(case3_gt_pixels, case3_contributions, "Case3")

            if weights is not None:
                case3_weights = weights[selections_case3]
                case3_contributions = case3_contributions * case3_weights[:, None, None]
            confusion_matrix += np.sum(case3_contributions, axis=0)

        ############################################################################################
        # Case 4: Both GT and preds contain labels that are missing from the other one
        #

        # [img_height, img_width]
        general_diff = gt ^ preds
        selections_case4 = np.logical_and(
            (general_diff & gt) > 0, (general_diff & preds) > 0
        )
        case4_preds_pixels = preds[selections_case4]
        case4_gt_pixels = gt[selections_case4]

        # [num_pixels_with_mutual_gt_pred_deltas, num_categories]
        case4_gt_preds_diff = (case4_preds_pixels ^ case4_gt_pixels) & case4_gt_pixels
        case4_gt_preds_diff = unpackbits(case4_gt_preds_diff, num_categories)

        # [num_pixels_with_mutual_gt_pred_deltas, num_categories]
        case4_preds_gt_diff = (
            case4_preds_pixels ^ case4_gt_pixels
        ) & case4_preds_pixels
        if len(case4_preds_gt_diff) > 0:
            case4_divider = np.bitwise_count(case4_preds_gt_diff)
            case4_preds_gt_diff = unpackbits(case4_preds_gt_diff, num_categories)

            # [num_pixels_with_mutual_gt_pred_deltas, num_categories]
            case4_preds_gt_intersection = case4_preds_pixels & case4_gt_pixels

            # [num_pixels_with_mutual_gt_pred_deltas, num_categories, num_categories]
            case4_preds_gt_intersection_diagonals = (
                unpackbits(case4_preds_gt_intersection, num_categories)[:, None, :]
                * eye[None, :, :]
            )

            # [num_pixels_with_mutual_gt_pred_deltas, num_categories, num_categories]
            case4_penalty = (
                case4_gt_preds_diff[:, :, None] * case4_preds_gt_diff[:, None, :]
            ) / case4_divider[:, None, None]
            case4_contributions = case4_penalty + case4_preds_gt_intersection_diagonals
            # print("Case4 contributions:")
            # print(case4_contributions)

            # Validate the contributions
            self._validate_contributions(case4_gt_pixels, case4_contributions, "Case4")

            if weights is not None:
                case4_weights = weights[selections_case4]
                case4_contributions = case4_contributions * case4_weights[:, None, None]
            confusion_matrix += np.sum(case4_contributions, axis=0)

        return confusion_matrix

    def compute_metrics(
        self,
        confusion_matrix: np.ndarray,
        class_names: dict[int, str],
    ) -> MultiLabelMatrixEvaluation:
        r"""
        Parameters:
        -----------
        confusion_matrix: np.ndarray[num_categories + 1, num_categories + 1]
        class_names: Mapping from class_id to class_names

        Returns
        --------

        """
        # Compute metrics on the full confusion matrix
        detailed_metrics = self._compute_matrix_metrics(confusion_matrix, class_names)

        # Collapse the classes except the background and compute metrics again
        collapsed_confusion_matrix = np.asarray(
            [
                [confusion_matrix[0, 0], np.sum(confusion_matrix[0, 1:])],
                [np.sum(confusion_matrix[1:, 0]), np.sum(confusion_matrix[1:, 1:])],
            ]
        )
        collapsed_class_names = {
            0: class_names[0],
            1: MultiLabelConfusionMatrix.ALL_COLLAPSED_CLASSES_NAME,
        }
        collapsed_metrics = self._compute_matrix_metrics(
            collapsed_confusion_matrix,
            collapsed_class_names,
        )

        evaluation = MultiLabelMatrixEvaluation(
            detailed=detailed_metrics, collapsed=collapsed_metrics
        )

        return evaluation

    def _compute_matrix_metrics(
        self,
        confusion_matrix: np.ndarray,
        class_names: dict[int, str],
    ) -> MultiLabelMatrixMetrics:
        r""" """
        col_sums = np.sum(confusion_matrix, axis=0)
        row_sums = np.sum(confusion_matrix, axis=1)

        # Compute precision_matrix and recall_matrix
        precision_matrix = np.divide(
            confusion_matrix,
            col_sums[None, :],
            out=np.zeros(confusion_matrix.shape),
            where=col_sums[None, :] != 0,
        )
        recall_matrix = np.divide(
            confusion_matrix,
            row_sums[:, None],
            out=np.zeros(confusion_matrix.shape),
            where=row_sums[:, None] != 0,
        )
        # Compute the f1 matrix element-wise
        f1_matrix_nom = 2 * precision_matrix * recall_matrix
        f1_matrix_denom = precision_matrix + recall_matrix
        f1_matrix = np.divide(
            f1_matrix_nom,
            f1_matrix_denom,
            out=np.zeros(confusion_matrix.shape),
            where=f1_matrix_denom != 0,
        )

        # Extract diagonal vectors
        precision = np.diag(precision_matrix)
        recall = np.diag(recall_matrix)
        f1 = np.diag(f1_matrix)
        precision_mean = np.average(precision)
        recall_mean = np.average(recall)
        f1_mean = np.average(f1)

        # Generate dicts with metrics per class name
        def get_class_name(class_id: int) -> str:
            return class_names[class_id]

        def array_to_dict(a: np.ndarray) -> dict[str, float]:
            a_dict = {get_class_name(i): float(x) for i, x in enumerate(a)}
            return a_dict

        precision_dict = array_to_dict(precision)
        recall_dict = array_to_dict(recall)
        f1_dict = array_to_dict(f1)

        agg_metrics = MultiLabelMatrixAggMetrics(
            classes_precision=precision_dict,
            classes_recall=recall_dict,
            classes_f1=f1_dict,
            classes_precision_mean=float(precision_mean),
            classes_recall_mean=float(recall_mean),
            classes_f1_mean=float(f1_mean),
        )

        metrics = MultiLabelMatrixMetrics(
            class_names=class_names,
            confusion_matrix=confusion_matrix,
            precision_matrix=precision_matrix,
            recall_matrix=recall_matrix,
            f1_matrix=f1_matrix,
            agg_metrics=agg_metrics,
        )
        return metrics

    def _validate_contributions(
        self,
        selected_gt: np.ndarray,
        contributions: np.ndarray,
        info: str,
    ):
        r"""
        Each contribution has the properties:
        1. The sum of each row corresponding to labels in GT is equal to one.
        2. The sum of all elements is equal to cardinality of GT.

        The validation is controled by self._validation_mode:
        - "disabled": No validation
        - "raise": Raise a ValueError
        - "log": Write an error log message

        Parameters:
        -----------
        selected_gt: np.ndarray  1D array with size=selected_pixels, each pixel is a uint64 encoding
        contributions: np.ndarray  [selected_pixels, num_classes, num_classes]
        """
        if self._validation_mode == "disabled":
            return

        contributions_shape = contributions.shape
        if len(contributions_shape) != 3:
            return

        num_categories = contributions_shape[1]

        selected_pixels = np.prod(selected_gt.shape)
        if selected_pixels != contributions_shape[0]:
            self._handle_error(f"{info}: Wrong contributions dimension")

        # Row sum check
        row_sum = np.sum(contributions, axis=2)
        expected_row_sum = unpackbits(selected_gt, num_categories)
        if not np.all(row_sum == expected_row_sum):
            self._handle_error(f"{info}: Wrong contributions row sums")

        # Full sum check
        full_sum: np.floating[Any] = np.sum(row_sum)
        expected_full_sum: np.uint64 = np.sum(np.bitwise_count(selected_gt))
        if full_sum != expected_full_sum:
            self._handle_error(f"{info}: Wrong contributions full sums")

    def _handle_error(self, msg: str):
        if self._validation_mode == "raise":
            raise ValueError(msg)
        else:
            _log.error(msg)



import logging
import random
import statistics
from pathlib import Path
from typing import List, Tuple

import matplotlib.pyplot as plt
import numpy as np

from docling_metrics_layout.layout_types import DatasetStatistics

_log = logging.getLogger(__name__)


def stats_to_table(
    dataset_stats: DatasetStatistics, metric_name: str
) -> Tuple[List[List[str]], List[str]]:
    """Convert dataset statistics to a formatted table."""
    headers = [
        f"{metric_name}",
        "prob [%]",
        "acc [%]",
        "1-acc [%]",
        "total",
    ]
    cumsum: float = 0.0

    table = []
    if dataset_stats.total > 0:
        for i in range(len(dataset_stats.bins) - 1):
            table.append(
                [
                    f"({dataset_stats.bins[i + 0]:.3f}, {dataset_stats.bins[i + 1]:.3f}]",
                    f"{100.0 * float(dataset_stats.hist[i]) / float(dataset_stats.total):.2f}",
                    f"{100.0 * cumsum:.2f}",
                    f"{100.0 * (1.0 - cumsum):.2f}",
                    f"{dataset_stats.hist[i]}",
                ]
            )
            cumsum += float(dataset_stats.hist[i]) / float(dataset_stats.total)
    return table, headers


def stats_to_histogram(dataset_stats: DatasetStatistics, figname: Path, name: str = ""):
    """Save dataset statistics as a histogram figure."""
    # Calculate bin widths
    bin_widths = [
        dataset_stats.bins[i + 1] - dataset_stats.bins[i]
        for i in range(len(dataset_stats.bins) - 1)
    ]
    bin_middle = [
        (dataset_stats.bins[i + 1] + dataset_stats.bins[i]) / 2.0
        for i in range(len(dataset_stats.bins) - 1)
    ]

    # Plot histogram
    fignum = int(1000 * random.random())
    plt.figure(fignum)
    plt.bar(bin_middle, dataset_stats.hist, width=bin_widths, edgecolor="black")

    plt.xlabel("Score")
    plt.ylabel("Frequency")
    plt.title(
        f"{name} (mean: {dataset_stats.mean:.2f}, median: {dataset_stats.median:.2f}, std: {dataset_stats.std:.2f}, total: {dataset_stats.total})"
    )

    _log.info(f"Saving figure to {figname}")
    plt.savefig(figname)


def compute_stats(
    values: List[float], max_value_is_one: bool = True, nr_bins: int = 20
) -> DatasetStatistics:
    total: int = len(values)

    mean: float = statistics.mean(values) if len(values) > 0 else -1
    median: float = statistics.median(values) if len(values) > 0 else -1
    std: float = statistics.stdev(values) if len(values) > 1 else 0.0
    _log.debug(
        f"Compute statistics: total: {total}, mean: {mean}, median: {median}, std: {std}"
    )

    max_value = 1.0
    if not max_value_is_one and len(values) > 0:
        max_value = max(values)

    # Compute the histogram
    hist, bins = np.histogram(values, bins=nr_bins, range=(0, max_value))
    _log.debug(f"Compute statistics: hist: {len(hist)}, #-bins: {len(bins)}")

    return DatasetStatistics(
        total=total, mean=mean, median=median, std=std, hist=hist, bins=bins
    )



import json
import tempfile
from pathlib import Path

from docling_metrics_layout.layout_types import (
    BboxResolution,
    DatasetToreLayoutEvaluation,
    LayoutMetricSample,
    MultiLabelMatrixEvaluation,
    PageToreEvaluation,
)
from docling_metrics_layout.tore.tore_layout_evaluator import (
    ToreLayoutEvaluator,
)

# Get the directory of this test file
TEST_DATA_DIR = Path(__file__).parent / "data"

# Tolerance for floating-point comparisons
FLOATING_POINT_TOLERANCE = 1e-6


def test_pixel_layout_evaluator():
    r"""Test LayoutMetrics evaluation on sample and dataset level."""
    # Load test data
    test_data_path = TEST_DATA_DIR / "dlnv1_t1_preds_score.json"
    with open(test_data_path) as f:
        test_data = json.load(f)

    # Create category mapping (using all unique category IDs found in test data)
    category_ids = set()
    for bbox in test_data["page_resolution_a"] + test_data["page_resolution_b"]:
        category_ids.add(bbox["category_id"])
    category_id_to_name = {cid: f"category_{cid}" for cid in sorted(category_ids)}

    # Create LayoutMetricSample from test data
    sample = LayoutMetricSample(
        id=test_data["id"],
        page_width=test_data["page_width"],
        page_height=test_data["page_height"],
        page_resolution_a=[
            BboxResolution(
                category_id=bbox["category_id"],
                bbox=bbox["bbox"],
            )
            for bbox in test_data["page_resolution_a"]
        ],
        page_resolution_b=[
            BboxResolution(
                category_id=bbox["category_id"],
                bbox=bbox["bbox"],
            )
            for bbox in test_data["page_resolution_b"]
        ],
    )

    # Test 1: Initialize LayoutMetrics and evaluate single sample
    evaluator = ToreLayoutEvaluator(
        category_id_to_name=category_id_to_name,
        concurrency=4,
    )

    sample_result = evaluator.evaluate_sample(sample)

    # Verify sample result structure
    assert isinstance(sample_result, PageToreEvaluation), (
        "sample_result must be PagePixelLayoutEvaluation instance"
    )
    assert sample_result.id == test_data["id"], (
        f"sample_result.id should be {test_data['id']}, got {sample_result.id}"
    )
    assert sample_result.num_pixels > 0, "num_pixels should be greater than 0"
    assert (
        sample_result.num_pixels == test_data["page_width"] * test_data["page_height"]
    ), "num_pixels should equal page_width * page_height"
    assert isinstance(
        sample_result.matrix_evaluation,
        MultiLabelMatrixEvaluation,
    ), "matrix_evaluation should be MultiLabelMatrixEvaluation instance"

    # Verify detailed metrics
    detailed_metrics = sample_result.matrix_evaluation.detailed
    assert detailed_metrics.confusion_matrix.shape[0] == 6, (
        f"detailed confusion matrix should have 6 classes, got {detailed_metrics.confusion_matrix.shape[0]}"
    )
    assert detailed_metrics.confusion_matrix.shape == (6, 6), (
        f"detailed confusion matrix shape should be (6, 6), got {detailed_metrics.confusion_matrix.shape}"
    )
    assert detailed_metrics.precision_matrix.shape == (6, 6), (
        f"detailed precision matrix shape should be (6, 6), got {detailed_metrics.precision_matrix.shape}"
    )
    assert detailed_metrics.recall_matrix.shape == (6, 6), (
        f"detailed recall matrix shape should be (6, 6), got {detailed_metrics.recall_matrix.shape}"
    )
    assert detailed_metrics.f1_matrix.shape == (6, 6), (
        f"detailed f1 matrix shape should be (6, 6), got {detailed_metrics.f1_matrix.shape}"
    )

    # Verify expected classes in detailed metrics
    expected_classes = {
        "Background",
        "category_0",
        "category_4",
        "category_6",
        "category_7",
        "category_9",
    }
    actual_classes = set(detailed_metrics.class_names.values())
    assert actual_classes == expected_classes, (
        f"detailed classes should be {expected_classes}, got {actual_classes}"
    )

    # Verify detailed metrics aggregations
    assert (
        abs(detailed_metrics.agg_metrics.classes_precision_mean - 0.3333333333333333)
        < FLOATING_POINT_TOLERANCE
    ), (
        f"detailed precision mean should be ~0.333, got {detailed_metrics.agg_metrics.classes_precision_mean}"
    )
    assert (
        abs(detailed_metrics.agg_metrics.classes_recall_mean - 0.3333333333333333)
        < FLOATING_POINT_TOLERANCE
    ), (
        f"detailed recall mean should be ~0.333, got {detailed_metrics.agg_metrics.classes_recall_mean}"
    )
    assert (
        abs(detailed_metrics.agg_metrics.classes_f1_mean - 0.3333333333333333)
        < FLOATING_POINT_TOLERANCE
    ), (
        f"detailed f1 mean should be ~0.333, got {detailed_metrics.agg_metrics.classes_f1_mean}"
    )

    # Verify collapsed metrics
    collapsed_metrics = sample_result.matrix_evaluation.collapsed
    assert collapsed_metrics.confusion_matrix.shape == (2, 2), (
        f"collapsed confusion matrix shape should be (2, 2), got {collapsed_metrics.confusion_matrix.shape}"
    )
    assert collapsed_metrics.precision_matrix.shape == (2, 2), (
        f"collapsed precision matrix shape should be (2, 2), got {collapsed_metrics.precision_matrix.shape}"
    )
    assert collapsed_metrics.recall_matrix.shape == (2, 2), (
        f"collapsed recall matrix shape should be (2, 2), got {collapsed_metrics.recall_matrix.shape}"
    )
    assert collapsed_metrics.f1_matrix.shape == (2, 2), (
        f"collapsed f1 matrix shape should be (2, 2), got {collapsed_metrics.f1_matrix.shape}"
    )

    # Verify expected classes in collapsed metrics
    expected_collapsed_classes = {"Background", "all_classes"}
    actual_collapsed_classes = set(collapsed_metrics.class_names.values())
    assert actual_collapsed_classes == expected_collapsed_classes, (
        f"collapsed classes should be {expected_collapsed_classes}, got {actual_collapsed_classes}"
    )

    # Verify collapsed metrics aggregations (all 1.0 for perfect match)
    assert collapsed_metrics.agg_metrics.classes_precision_mean == 1.0, (
        f"collapsed precision mean should be 1.0, got {collapsed_metrics.agg_metrics.classes_precision_mean}"
    )
    assert collapsed_metrics.agg_metrics.classes_recall_mean == 1.0, (
        f"collapsed recall mean should be 1.0, got {collapsed_metrics.agg_metrics.classes_recall_mean}"
    )
    assert collapsed_metrics.agg_metrics.classes_f1_mean == 1.0, (
        f"collapsed f1 mean should be 1.0, got {collapsed_metrics.agg_metrics.classes_f1_mean}"
    )

    # Test 2: Evaluate dataset (single sample in this case)
    dataset_result = evaluator.evaluate_dataset([sample])

    # Verify dataset result structure
    assert isinstance(dataset_result, DatasetToreLayoutEvaluation), (
        "dataset_result must be DatasetPixelLayoutEvaluation instance"
    )
    assert dataset_result.num_pages == 1, (
        f"num_pages should be 1, got {dataset_result.num_pages}"
    )
    assert dataset_result.num_pixels > 0, "num_pixels should be greater than 0"
    assert (
        dataset_result.num_pixels == test_data["page_width"] * test_data["page_height"]
    ), "num_pixels should equal page_width * page_height"
    assert test_data["id"] in dataset_result.page_evaluations, (
        f"page_evaluations should contain {test_data['id']}"
    )
    assert len(dataset_result.page_evaluations) == 1, (
        f"page_evaluations should have 1 entry, got {len(dataset_result.page_evaluations)}"
    )

    # Test 3: Evaluate dataset with save functionality
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_root = Path(tmp_dir)

        # Evaluate and export
        dataset_result_with_save = evaluator.evaluate_dataset([sample])
        evaluator.export_evaluations(dataset_result_with_save, tmp_root)

        # Verify dataset result is created
        assert isinstance(dataset_result_with_save, DatasetToreLayoutEvaluation), (
            "dataset_result_with_save must be DatasetPixelLayoutEvaluation instance"
        )
        assert dataset_result_with_save.num_pages == 1, (
            f"num_pages should be 1, got {dataset_result_with_save.num_pages}"
        )

        # Verify the expected reports from PixelLayoutEvaluator
        report_filenames = ToreLayoutEvaluator.evaluation_filenames(tmp_root)
        excel_report_fn = report_filenames["excel"]
        assert excel_report_fn.is_file(), f"Missing report at {excel_report_fn}"



import logging
from pathlib import Path

from docling_metrics_layout import (
    LayoutMetrics,
    LayoutMetricSample,
)
from docling_metrics_layout.layout_types import LayoutMetricDatasetEvaluation

# Data
DATA = {
    "id": "page_001",
    "page_width": 612,
    "page_height": 792,
    "page_resolution_a": [
        {"category_id": 1, "bbox": [50.0, 80.0, 300.0, 200.0], "score": 0.97},
        {"category_id": 2, "bbox": [310.0, 80.0, 560.0, 200.0], "score": 0.91},
        {"category_id": 1, "bbox": [50.0, 220.0, 560.0, 380.0], "score": 0.88},
        {"category_id": 3, "bbox": [50.0, 400.0, 300.0, 500.0], "score": 0.76},
    ],
    "page_resolution_b": [
        {"category_id": 1, "bbox": [52.0, 82.0, 298.0, 202.0], "score": 0.95},
        {"category_id": 2, "bbox": [312.0, 78.0, 558.0, 199.0], "score": 0.89},
        {"category_id": 1, "bbox": [48.0, 218.0, 562.0, 382.0], "score": 0.85},
    ],
}

CATEGORIES = {1: "red", 2: "orange", 3: "green", 4: "blue"}

REPORT_DIR = "reports/"


def setup_logging():
    log_format = "%(asctime)s - %(levelname)s - %(message)s"
    logging.basicConfig(level=logging.INFO, format=log_format)


def demo():
    # Load the data as list[LayoutMetricSample]
    samples = [LayoutMetricSample.model_validate(DATA)]

    # Evaluate
    lm = LayoutMetrics(CATEGORIES, save_root=Path(REPORT_DIR))
    evaluation: LayoutMetricDatasetEvaluation = lm.evaluate_dataset(samples)

    # mAP metrics
    map_eval = evaluation.dataset_map_layout_evaluation
    print(f"\nmAP:         {map_eval.map:.4f}")
    print(f"mAP per class: {map_eval.map_per_class}")

    # TORE metrics
    tore_matrix = evaluation.dataset_tore_evaluation.matrix_evaluation.detailed
    print(f"\nPrecision matrix:\n{tore_matrix.precision_matrix}")
    print(f"\nRecall matrix:\n{tore_matrix.recall_matrix}")

    # Report files
    print("\nReport files:")
    for f in evaluation.reports:
        print(f"  {f.name}")


if __name__ == "__main__":
    setup_logging()
    demo()



import json
import logging
from pathlib import Path

from docling_metrics_layout.layout_types import BboxResolution, LayoutMetricSample
from docling_metrics_layout.utils.utils import xywh_to_xyxy

_log = logging.getLogger(__name__)


def load_layout_samples(
    gt_coco_fn: Path,
    preds_coco_fn: Path,
) -> tuple[dict[int, str], list[LayoutMetricSample]]:
    r"""
    Load LayoutMetricSample objects from COCO GT and predictions files.

    Parameters:
        gt_coco_fn: Path to COCO ground-truth annotations JSON file.
        preds_coco_fn: Path to COCO predictions JSON file.

    Returns:
        Tuple of (category_id_to_name mapping, list of LayoutMetricSample).
    """
    with open(gt_coco_fn) as f:
        gt_coco = json.load(f)

    with open(preds_coco_fn) as f:
        predictions = json.load(f)

    # Build category mapping
    category_id_to_name: dict[int, str] = {
        cat["id"]: cat["name"] for cat in gt_coco["categories"]
    }

    # Build image info mapping: image_id -> image dict (includes width, height)
    image_info: dict[int, dict] = {img["id"]: img for img in gt_coco["images"]}

    # Build GT bboxes per image: image_id -> list[BboxResolution]
    gt_per_image: dict[int, list[BboxResolution]] = {}
    for ann in gt_coco["annotations"]:
        image_id = ann["image_id"]
        gt_per_image.setdefault(image_id, []).append(
            BboxResolution(
                category_id=ann["category_id"],
                bbox=xywh_to_xyxy(ann["bbox"]),
            )
        )

    # Build prediction bboxes per image: image_id -> list[BboxResolution]
    preds_per_image: dict[int, list[BboxResolution]] = {}
    for pred in predictions:
        image_id = pred["image_id"]
        preds_per_image.setdefault(image_id, []).append(
            BboxResolution(
                category_id=pred["category_id"],
                bbox=xywh_to_xyxy(pred["bbox"]),
                score=pred.get("score"),
            )
        )

    # Build one LayoutMetricSample per image that has predictions
    samples: list[LayoutMetricSample] = []
    for image_id, img in image_info.items():
        gt_bboxes = gt_per_image.get(image_id, [])
        pred_bboxes = preds_per_image.get(image_id)
        if pred_bboxes is None:
            _log.warning("Missing predictions for image_id: %s", image_id)
            continue
        sample = LayoutMetricSample(
            id=str(image_id),
            page_width=img["width"],
            page_height=img["height"],
            page_resolution_a=gt_bboxes,
            page_resolution_b=pred_bboxes,
        )
        samples.append(sample)

    _log.info("Loaded %d samples", len(samples))
    return category_id_to_name, samples



import colorsys
import logging
from pathlib import Path
from typing import Any, Optional

import numpy as np
import pandas as pd
from openpyxl import Workbook
from openpyxl.styles import Border, Font, PatternFill, Side
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.worksheet import Worksheet
from pandas import ExcelWriter

from docling_metrics_layout.layout_types import (
    MultiLabelMatrixEvaluation,
)
from docling_metrics_layout.tore.multi_label_confusion_matrix import (
    MultiLabelConfusionMatrix,
)

_log = logging.getLogger(__name__)


def linear_norm(x, x_min, x_max, k=5.0):
    d = x_max - x_min
    x = (x - x_min) / (x_max - x_min) if d != 0 else 0
    return x


def exp_norm(x, x_min, x_max, k=5.0):
    d = x_max - x_min
    x = (x - x_min) / (x_max - x_min) if d != 0 else 0
    return 1 - np.exp(-k * x)


def power_norm(x, x_min, x_max, p=0.3):
    d = x_max - x_min
    x = (x - x_min) / (x_max - x_min) if d != 0 else 0
    return x**p


def discover_filename_prefix(
    root: Path, filename_extension: str, scan_depth: int = 10
) -> Optional[str]:
    r"""
    Discover the common prefix used in the filenames of the prediction visualisations
    Check up to scan_depth files in the dir

    Return
    -------
    empty string: There is no common prefix
    None: Cannot reach consensus in any common prefix
    """

    def common_prefix(a: str, b: str, stop_char="_"):
        r"""String common prefix with stop char"""
        prefix: list[str] = []
        for c1, c2 in zip(a, b):
            if c1 != c2:
                break
            prefix.append(c1)
            if c1 == stop_char:
                break
        return "".join(prefix)

    prev_image_filename = None
    prefix = ""
    for i, image_fn in enumerate(root.glob(f"*.{filename_extension}")):
        if i >= scan_depth:
            break
        image_filename = image_fn.name
        if prev_image_filename:
            # new_prefix will be the empty string if there is nothing in common
            new_prefix = common_prefix(image_filename, prev_image_filename)
            if prefix == "":
                prefix = new_prefix
            elif new_prefix != "" and prefix != new_prefix:
                return None
        prev_image_filename = image_filename
    return prefix


class ConfusionMatrixExporter:
    r""" """

    TITLE_FONT_SIZE = 16
    SUBTITLE_FONT_SIZE = 14
    DATASET_WORKSHEET_NAME = "Dataset"
    IMAGES_WORKSHEET_NAME = "Images"

    def __init__(
        self,
    ):
        self._background_color = "bbbbbb"
        self._black_color = "444444"
        self._power_normalization_exp = 0.2

        border_style = "thick"
        border_color = "fc5a8d"
        self._highlighted_border = Border(
            left=Side(border_style=border_style, color=border_color),
            right=Side(border_style=border_style, color=border_color),
            top=Side(border_style=border_style, color=border_color),
            bottom=Side(border_style=border_style, color=border_color),
        )

    def build_ds_report(
        self,
        title: str,
        num_images: int,
        num_pixels: int,
        headers: list[str],
        matrix_evaluation: MultiLabelMatrixEvaluation,
        collapsed_headers: list[str],
        image_collaped_aggs: dict[str, np.ndarray],
        excel_fn: Path,
        visualisations_root: Optional[Path] = None,
    ):
        r"""
        Generate excel report for the full dataset
        """
        with pd.ExcelWriter(excel_fn, engine="openpyxl") as writer:
            # Add the dataset header
            wb: Workbook = writer.book
            if not wb.worksheets:
                wb.create_sheet(ConfusionMatrixExporter.DATASET_WORKSHEET_NAME)
                wb.active = 0
            ds_ws: Worksheet = wb.active  # type: ignore
            ds_ws.cell(row=1, column=1).value = title
            ds_ws.cell(row=1, column=1).font = Font(
                bold=True, size=ConfusionMatrixExporter.TITLE_FONT_SIZE
            )
            ds_ws.cell(row=2, column=1).value = "#images"
            ds_ws.cell(row=2, column=2).value = num_images
            ds_ws.cell(row=3, column=1).value = "#pixels"
            ds_ws.cell(row=3, column=2).value = num_pixels
            ds_ws.cell(row=3, column=2).number_format = "#,##0"

            # Build the basic report
            self._build_base_report(
                writer,
                ConfusionMatrixExporter.DATASET_WORKSHEET_NAME,
                headers,
                matrix_evaluation,
                4,
            )

            # Add the collapsed image metrics in a separate worksheet
            self._aggregate_collapsed_image_metrics(
                writer,
                ConfusionMatrixExporter.IMAGES_WORKSHEET_NAME,
                collapsed_headers,
                image_collaped_aggs,
                visualisations_root=visualisations_root,
            )

            # Adjust column widths
            images_ws: Worksheet = wb[ConfusionMatrixExporter.IMAGES_WORKSHEET_NAME]
            self._adjust_column_widths(ds_ws)
            self._adjust_column_widths(images_ws)

        _log.info("Dataset report: %s", str(excel_fn))

    def build_image_report(
        self,
        headers: list[str],
        matrix_evaluation: MultiLabelMatrixEvaluation,
        excel_fn: Path,
    ):
        with pd.ExcelWriter(excel_fn, engine="openpyxl") as writer:
            self._build_base_report(
                writer,
                ConfusionMatrixExporter.DATASET_WORKSHEET_NAME,
                headers,
                matrix_evaluation,
            )

            # Adjust column widths
            ws: Worksheet = writer.book[ConfusionMatrixExporter.DATASET_WORKSHEET_NAME]
            self._adjust_column_widths(ws)

        _log.info("Image report: %s", str(excel_fn))

    def _aggregate_collapsed_image_metrics(
        self,
        writer: ExcelWriter,
        worksheet_name: str,
        headers: list[str],
        image_collapsed_aggs: dict[str, np.ndarray],
        origin_cell: tuple[int, int] = (0, 0),
        decimal_digits: int = 3,
        visualisations_root: Optional[Path] = None,
    ):
        r"""
        Aggregate all collapsed image metrics
        """
        startrow = origin_cell[0] + 1
        startcol = origin_cell[1]

        # Build the dataframe
        index = list(image_collapsed_aggs.keys())
        data = np.stack(list(image_collapsed_aggs.values()), axis=0)  # [num_images, 12]
        data = np.round(data, decimals=3)
        df = pd.DataFrame(data, index=index, columns=headers)

        df.to_excel(
            writer,
            sheet_name=worksheet_name,
            index=True,
            startrow=startrow,
            startcol=startcol,
        )  # row/col index starts from 0

        # Load workbook
        wb: Workbook = writer.book
        ws: Worksheet = wb[worksheet_name]

        # Set the prediction visualisations as hyperlinks in the image filenames
        if visualisations_root:
            viz_prefix = discover_filename_prefix(visualisations_root, "png")
            if viz_prefix:
                col = startcol + 1
                for i, image_filename in enumerate(image_collapsed_aggs.keys()):
                    row = i + startrow + 2
                    cell = ws.cell(row=row, column=col)
                    viz_fn = visualisations_root / f"{viz_prefix}{image_filename}"
                    if not viz_fn.is_file():
                        continue
                    cell.hyperlink = str(viz_fn)
                    cell.style = "Hyperlink"
            else:
                _log.error(
                    "Cannot find any visualisation prefix in: %s",
                    str(visualisations_root),
                )

        # Set the subtitle
        subtitle_cell = ws.cell(
            row=origin_cell[0] + 1, column=origin_cell[1] + 1
        )  # start from 1
        subtitle_cell.value = "Image collapsed classes metrics"
        subtitle_cell.font = Font(
            bold=True, size=ConfusionMatrixExporter.SUBTITLE_FONT_SIZE
        )

        # Get data min/max values
        vmin: np.floating[Any] = np.min(data)
        vmax: np.floating[Any] = np.max(data)

        # Apply background colors to data cells
        row_start = 2 + startrow  # start from 1
        col_start = 2 + startcol
        for i in range(data.shape[0]):
            row = i + row_start
            for j in range(data.shape[1]):
                value = data[i, j]
                col = j + col_start
                if value == 0:
                    # Treat zero values specially
                    color = self._black_color
                else:
                    color = self._value_to_color(vmin, vmax, value, "linear")
                ws.cell(row=row, column=col).fill = PatternFill(
                    start_color=color, end_color=color, fill_type="solid"
                )

                # Format the numbers
                decimals_format = "." + "0" * decimal_digits
                ws.cell(row=row, column=col).number_format = f"#,##0{decimals_format}"
        return ws.max_row, ws.max_column

    def _build_base_report(
        self,
        writer: ExcelWriter,
        worksheet_name: str,
        headers: list[str],
        matrix_evaluation: MultiLabelMatrixEvaluation,
        startrow: int = 0,
        hide_zero_rows: bool = True,
        hide_zero_cols: bool = True,
    ):
        r"""
        Generate excel report for a single image
        """
        detailed_spacing = 4  # spacing between the detailed matrices
        collapsed_spacing = (
            2  # spacing between a detailed and the next collapsed matrix
        )

        collapsed_headers = [
            headers[0],
            MultiLabelConfusionMatrix.ALL_COLLAPSED_CLASSES_NAME,
        ]

        # Add the confusion matrix
        max_row, max_col = self._export_matrix_to_excel(
            writer,
            worksheet_name,
            "Confusion Matrix",
            matrix_evaluation.detailed.confusion_matrix,
            headers,
            decimal_digits=3,
            origin_cell=(startrow, 0),
            normalization_func="linear",
            hide_zero_rows=hide_zero_rows,
            hide_zero_cols=hide_zero_cols,
        )

        # Add the precision matrix with detailed classes
        detailed_precision_row = max_row + detailed_spacing
        collapsed_precision_row = max_row + collapsed_spacing
        max_row, max_col = self._export_matrix_to_excel(
            writer,
            worksheet_name,
            "Precision Matrix",
            matrix_evaluation.detailed.precision_matrix,
            headers,
            decimal_digits=3,
            origin_cell=(detailed_precision_row, 0),
            normalization_func="linear",
            hide_zero_rows=hide_zero_rows,
            hide_zero_cols=hide_zero_cols,
        )
        detailed_recall_row = max_row + detailed_spacing
        collapsed_recall_row = max_row + collapsed_spacing
        collapsed_col = max_col + 1

        # Add the precision matrix with collapsed classes
        self._export_matrix_to_excel(
            writer,
            worksheet_name,
            "Collapsed Precision Matrix",
            matrix_evaluation.collapsed.precision_matrix,
            collapsed_headers,
            decimal_digits=3,
            origin_cell=(collapsed_precision_row, collapsed_col),
            normalization_func="linear",
            hide_zero_rows=hide_zero_rows,
            hide_zero_cols=hide_zero_cols,
        )

        # Add the recall matrix with detailed classes
        max_row, max_col = self._export_matrix_to_excel(
            writer,
            worksheet_name,
            "Recall matrix",
            matrix_evaluation.detailed.recall_matrix,
            headers,
            decimal_digits=3,
            origin_cell=(detailed_recall_row, 0),
            normalization_func="linear",
            hide_zero_rows=hide_zero_rows,
            hide_zero_cols=hide_zero_cols,
        )

        # Add the recall matrix with collapsed classes
        self._export_matrix_to_excel(
            writer,
            worksheet_name,
            "Collapsed Recall Matrix",
            matrix_evaluation.collapsed.recall_matrix,
            collapsed_headers,
            decimal_digits=3,
            origin_cell=(collapsed_recall_row, collapsed_col),
            normalization_func="linear",
            hide_zero_rows=hide_zero_rows,
            hide_zero_cols=hide_zero_cols,
        )

    def _export_matrix_to_excel(
        self,
        writer: ExcelWriter,
        worksheet_name: str,
        title: str,
        data: np.ndarray,
        headers: list[str],
        origin_cell: tuple[int, int] = (0, 0),
        special_first_cell: bool = False,
        decimal_digits: int = 0,
        normalization_func: str = "linear",  # One of 'linear', 'power', 'exp'
        hide_zero_rows: bool = False,
        hide_zero_cols: bool = False,
    ) -> tuple[int, int]:
        r"""
        Export the given data in excel and place it in the origin_cell

        Returns:
        --------
        max_row
        max_col
        """
        startrow = origin_cell[0]
        startcol = origin_cell[1]

        # Round values
        data = np.round(data, decimals=3)

        # Create DataFrame and write to Excel
        df = pd.DataFrame(data, index=headers, columns=headers)
        df.to_excel(
            writer,
            sheet_name=worksheet_name,
            index=True,
            startrow=startrow,
            startcol=startcol,
        )  # row/col index starts from 0

        # Load workbook
        wb: Workbook = writer.book
        ws: Worksheet = wb[worksheet_name]

        # Set the subtitle in the corner of the data
        subtitle_cell = ws.cell(
            row=origin_cell[0] + 1, column=origin_cell[1] + 1
        )  # start from 1
        subtitle_cell.value = title
        subtitle_cell.font = Font(
            bold=True, size=ConfusionMatrixExporter.SUBTITLE_FONT_SIZE
        )

        # Get the min/max values
        if special_first_cell:
            # Don't account for the first value
            confusion_mask = np.ones(data.shape, dtype=np.uint8)
            confusion_mask[0, 0] = 0
            vmin: np.floating[Any] = np.min(data, initial=0, where=confusion_mask != 0)
            vmax: np.floating[Any] = np.max(data, initial=0, where=confusion_mask != 0)
        else:
            vmin = np.min(data)
            vmax = np.max(data)

        # Apply background colors to data cells
        style_startrow = 2 + startrow  # start from 1
        style_startcol = 2 + startcol
        for i in range(len(headers)):
            row = i + style_startrow
            for j in range(len(headers)):
                col = j + style_startcol
                # Treat the background specially
                if i == 0 and j == 0 and special_first_cell:
                    ws.cell(row=row, column=col).fill = PatternFill(
                        start_color=self._background_color,
                        end_color=self._background_color,
                        fill_type="solid",
                    )
                else:
                    value = data[i, j]
                    if value == 0:
                        # Treat zero values specially
                        color = self._black_color
                    else:
                        color = self._value_to_color(
                            vmin, vmax, value, normalization_func
                        )
                    ws.cell(row=row, column=col).fill = PatternFill(
                        start_color=color, end_color=color, fill_type="solid"
                    )

                # Highlight the diagonal
                if i == j:
                    ws.cell(row=row, column=col).border = self._highlighted_border

                # Format the numbers
                decimals_format = ""
                if decimal_digits > 0:
                    decimals_format = "." + "0" * decimal_digits
                ws.cell(row=row, column=col).number_format = f"#,##0{decimals_format}"

        # Hide rows/cols with all zeros
        if hide_zero_cols:
            colsums = np.sum(data, axis=0)
            zero_col_indices = np.nonzero(colsums == 0)[0]  # Zero column indices
            for zero_col_idx in zero_col_indices:
                col_idx = zero_col_idx + startcol + 2
                col_letter = get_column_letter(col_idx)
                ws.column_dimensions[col_letter].hidden = True

        if hide_zero_rows:
            rowsums = np.sum(data, axis=1)
            zero_row_indices = np.nonzero(rowsums == 0)[0]  # Zero row indices
            for zero_row_idx in zero_row_indices:
                row_idx = zero_row_idx + startrow + 2
                ws.row_dimensions[row_idx].hidden = True

        return ws.max_row, ws.max_column

    def _value_to_color(self, vmin, vmax, v, normalization_func: str):
        """Map value to RGB color from blue→red using rainbow spectrum."""
        # Normalize to [0,1]
        if normalization_func == "power":
            normalized_value = power_norm(
                v, vmin, vmax, p=self._power_normalization_exp
            )
        elif normalization_func == "exp":
            normalized_value = exp_norm(v, vmin, vmax)
        else:
            normalized_value = linear_norm(v, vmin, vmax)

        # Use HSV rainbow mapping: hue 240° (blue) -> 0° (red)
        hue = (1 - normalized_value) * 240 / 360  # convert degrees to [0,1]
        r, g, b = colorsys.hsv_to_rgb(hue, 1, 1)

        # Convert to hex color for Excel
        hex_color = f"{int(r * 255):02X}{int(g * 255):02X}{int(b * 255):02X}"
        return hex_color

    def _adjust_column_widths(self, ws: Worksheet):
        r"""Adjust column widths for the final excel"""
        for col in ws.columns:
            max_length = 0
            col_letter = col[0].column_letter  # type: ignore
            for cell in col:
                val = str(cell.value)
                max_length = max(max_length, len(val))
            ws.column_dimensions[col_letter].width = max_length + 2



from pathlib import Path
from typing import Any, Optional

import numpy as np
from docling_metrics_core.base_types import (
    BaseAggregateResult,
    BaseInputSample,
    BaseSampleResult,
)
from pydantic import BaseModel, model_serializer, model_validator


class DatasetStatistics(BaseModel):
    total: int

    mean: float
    median: float
    std: float

    bins: list[float]
    hist: list[float]

    @model_validator(mode="after")
    def check_bins_and_hist_lengths(self):
        if len(self.bins) != len(self.hist) + 1:
            raise ValueError("`bins` must have exactly one more element than `hist`.")
        return self


class BboxResolution(BaseModel):
    r"""Single bbox resolution"""

    category_id: int

    # bbox coords: (x1, y1, x2, y2) with the origin(0, 0) at the top, left corner, no normalization
    bbox: list[float]
    score: Optional[float] = None


class MultiLabelMatrixAggMetrics(BaseModel):
    classes_precision: dict[str, float]
    classes_recall: dict[str, float]
    classes_f1: dict[str, float]

    classes_precision_mean: float
    classes_recall_mean: float
    classes_f1_mean: float


class MultiLabelMatrixMetrics(BaseModel):
    model_config = {"arbitrary_types_allowed": True}

    class_names: dict[int, str]
    confusion_matrix: np.ndarray
    precision_matrix: np.ndarray
    recall_matrix: np.ndarray
    f1_matrix: np.ndarray

    agg_metrics: MultiLabelMatrixAggMetrics

    @model_serializer(mode="wrap")
    def serialize_model(self, serializer: Any) -> dict:
        data = serializer(self)
        for field_name, field_value in self.__dict__.items():
            if isinstance(field_value, np.ndarray):
                data[field_name] = field_value.tolist()
        return data

    @model_validator(mode="before")
    @classmethod
    def deserialize_arrays(cls, data: Any) -> Any:
        if isinstance(data, dict):
            array_fields = [
                "confusion_matrix",
                "precision_matrix",
                "recall_matrix",
                "f1_matrix",
            ]
            for field_name in array_fields:
                if field_name in data:
                    data[field_name] = np.asarray(data[field_name])
        return data


class MultiLabelMatrixEvaluation(BaseModel):
    detailed: MultiLabelMatrixMetrics  # All classes included
    collapsed: MultiLabelMatrixMetrics  # Only the background the other classes summed up together


class PageToreEvaluation(BaseModel):
    id: str
    num_pixels: int
    matrix_evaluation: MultiLabelMatrixEvaluation


class DatasetToreLayoutEvaluation(BaseModel):
    num_pages: int
    num_pixels: int
    matrix_evaluation: MultiLabelMatrixEvaluation
    page_evaluations: dict[str, PageToreEvaluation]


class MAPMetrics(BaseModel):
    map: float
    map_50: float
    map_75: float
    map_large: float
    map_medium: float
    map_per_class: dict[str, float]
    map_small: float
    mar_1: float
    mar_10: float
    mar_100: float
    mar_100_per_class: dict[str, float]
    mar_large: float
    mar_medium: float
    mar_small: float


class MAPPageLayoutEvaluation(MAPMetrics):
    id: str


class MAPDatasetLayoutEvaluation(MAPMetrics):
    page_evaluations: dict[str, MAPPageLayoutEvaluation]

    # Statistics across the corresponding mAP metrics across all dataset pages
    map_stats: DatasetStatistics
    map_50_stats: DatasetStatistics
    map_75_stats: DatasetStatistics


class LayoutMetricSample(BaseInputSample):
    page_width: int
    page_height: int
    page_resolution_a: list[BboxResolution]
    page_resolution_b: list[BboxResolution]


class LayoutMetricSampleEvaluation(BaseSampleResult):
    r"""Layout evaluation for one page"""

    page_tore_evaluation: PageToreEvaluation
    page_map_layout_evaluation: MAPPageLayoutEvaluation


class LayoutMetricDatasetEvaluation(BaseAggregateResult):
    r"""Layout evaluation for the entire dataset"""

    dataset_tore_evaluation: DatasetToreLayoutEvaluation
    dataset_map_layout_evaluation: MAPDatasetLayoutEvaluation
    reports: Optional[list[Path]] = None



from docling_metrics_layout.docling_metrics_layout import (
    LayoutMetrics,
)
from docling_metrics_layout.layout_types import (
    BboxResolution,
    DatasetToreLayoutEvaluation,
    LayoutMetricDatasetEvaluation,
    LayoutMetricSample,
    LayoutMetricSampleEvaluation,
    MultiLabelMatrixAggMetrics,
    MultiLabelMatrixEvaluation,
    MultiLabelMatrixMetrics,
    PageToreEvaluation,
)

__all__ = [
    "BboxResolution",
    "DatasetToreLayoutEvaluation",
    "LayoutMetricDatasetEvaluation",
    "LayoutMetricSample",
    "LayoutMetricSampleEvaluation",
    "LayoutMetrics",
    "MultiLabelMatrixAggMetrics",
    "MultiLabelMatrixEvaluation",
    "MultiLabelMatrixMetrics",
    "PageToreEvaluation",
]



import json
from pathlib import Path

from docling_metrics_layout.layout_types import (
    BboxResolution,
    LayoutMetricSample,
    MAPDatasetLayoutEvaluation,
    MAPPageLayoutEvaluation,
)
from docling_metrics_layout.map.map_layout_evaluator import (
    MAPLayoutEvaluator,
)

# Get the directory of this test file
TEST_DATA_DIR = Path(__file__).parent / "data"

# Tolerance for floating-point comparisons
FLOATING_POINT_TOLERANCE = 1e-6


def test_map_layout_evaluations():
    r"""Test MAPLayoutEvaluator evaluation on sample and dataset level with actual value assertions."""
    test_data_path = TEST_DATA_DIR / "dlnv1_t1_preds_score.json"

    # Load the test data with scores
    with open(test_data_path) as f:
        test_data = json.load(f)

    # Create category mapping (using all unique category IDs found in test data)
    category_ids = set()
    for bbox in test_data["page_resolution_a"] + test_data["page_resolution_b"]:
        category_ids.add(bbox["category_id"])
    category_id_to_name = {cid: f"category_{cid}" for cid in sorted(category_ids)}

    # Create LayoutMetricSample from test data with confidence scores
    sample = LayoutMetricSample(
        id=test_data["id"],
        page_width=test_data["page_width"],
        page_height=test_data["page_height"],
        page_resolution_a=[
            BboxResolution(
                category_id=bbox["category_id"],
                bbox=bbox["bbox"],
                score=bbox.get("score", 1.0),
            )
            for bbox in test_data["page_resolution_a"]
        ],
        page_resolution_b=[
            BboxResolution(
                category_id=bbox["category_id"],
                bbox=bbox["bbox"],
                score=bbox.get("score", 1.0),
            )
            for bbox in test_data["page_resolution_b"]
        ],
    )

    # Initialize MAPLayoutEvaluator
    evaluator = MAPLayoutEvaluator(
        category_id_to_name=category_id_to_name,
    )

    # Test 1: Evaluate single sample
    sample_result = evaluator.evaluate_sample(sample)

    # Verify sample result is correct type
    assert isinstance(sample_result, MAPPageLayoutEvaluation)
    assert sample_result.id == test_data["id"]

    # Verify primary map scores
    assert sample_result.map == 1.0, (
        f"map should be 1.0 (perfect match), got {sample_result.map}"
    )
    assert sample_result.map_50 == 1.0, (
        f"map_50 should be 1.0, got {sample_result.map_50}"
    )
    assert sample_result.map_75 == 1.0, (
        f"map_75 should be 1.0, got {sample_result.map_75}"
    )

    # Verify area-based map scores
    assert sample_result.map_small == 1.0, (
        f"map_small should be 1.0, got {sample_result.map_small}"
    )
    assert sample_result.map_medium == 1.0, (
        f"map_medium should be 1.0, got {sample_result.map_medium}"
    )
    assert sample_result.map_large == 1.0, (
        f"map_large should be 1.0, got {sample_result.map_large}"
    )

    # Verify recall scores
    assert abs(sample_result.mar_1 - 0.6) < 0.1, (
        f"mar_1 should be ~0.6, got {sample_result.mar_1}"
    )
    assert sample_result.mar_10 == 1.0, (
        f"mar_10 should be 1.0, got {sample_result.mar_10}"
    )
    assert sample_result.mar_100 == 1.0, (
        f"mar_100 should be 1.0, got {sample_result.mar_100}"
    )

    # Verify area-based recall scores
    assert sample_result.mar_small == 1.0, (
        f"mar_small should be 1.0, got {sample_result.mar_small}"
    )
    assert sample_result.mar_medium == 1.0, (
        f"mar_medium should be 1.0, got {sample_result.mar_medium}"
    )
    assert sample_result.mar_large == 1.0, (
        f"mar_large should be 1.0, got {sample_result.mar_large}"
    )

    # Verify per-class metrics
    expected_classes = {
        "category_0",
        "category_4",
        "category_6",
        "category_7",
        "category_9",
    }
    assert set(sample_result.map_per_class.keys()) == expected_classes, (
        f"map_per_class keys should be {expected_classes}, got {set(sample_result.map_per_class.keys())}"
    )
    assert all(v == 1.0 for v in sample_result.map_per_class.values()), (
        f"All per-class map values should be 1.0, got {sample_result.map_per_class}"
    )

    assert set(sample_result.mar_100_per_class.keys()) == expected_classes, (
        f"mar_100_per_class keys should be {expected_classes}, got {set(sample_result.mar_100_per_class.keys())}"
    )
    assert all(v == 1.0 for v in sample_result.mar_100_per_class.values()), (
        f"All per-class mar_100 values should be 1.0, got {sample_result.mar_100_per_class}"
    )

    # Test 2: Evaluate dataset
    dataset_result = evaluator.evaluate_dataset([sample])

    # Verify dataset result is correct type
    assert isinstance(dataset_result, MAPDatasetLayoutEvaluation)

    # Verify dataset-level map scores
    assert dataset_result.map == 1.0, (
        f"dataset map should be 1.0, got {dataset_result.map}"
    )
    assert dataset_result.map_50 == 1.0, (
        f"dataset map_50 should be 1.0, got {dataset_result.map_50}"
    )
    assert dataset_result.map_75 == 1.0, (
        f"dataset map_75 should be 1.0, got {dataset_result.map_75}"
    )

    # Verify page evaluations
    assert len(dataset_result.page_evaluations) == 1, (
        f"page_evaluations should have 1 entry, got {len(dataset_result.page_evaluations)}"
    )
    assert test_data["id"] in dataset_result.page_evaluations, (
        f"page_evaluations should contain {test_data['id']}"
    )

    # Verify page evaluation values match sample evaluation
    page_eval = dataset_result.page_evaluations[test_data["id"]]
    assert isinstance(page_eval, MAPPageLayoutEvaluation)
    assert page_eval.id == test_data["id"]
    assert page_eval.map == sample_result.map, (
        f"page_eval.map should match sample_result.map: {page_eval.map} vs {sample_result.map}"
    )
    assert page_eval.map_50 == sample_result.map_50, (
        f"page_eval.map_50 should match sample_result.map_50: {page_eval.map_50} vs {sample_result.map_50}"
    )

    # Verify dataset-level stats exist and are correct (single page → total=1, mean=median=page value, std=0)
    assert dataset_result.map_stats.total == 1
    assert (
        abs(dataset_result.map_stats.mean - sample_result.map)
        < FLOATING_POINT_TOLERANCE
    )
    assert (
        abs(dataset_result.map_stats.median - sample_result.map)
        < FLOATING_POINT_TOLERANCE
    )
    assert dataset_result.map_stats.std == 0.0

    assert dataset_result.map_50_stats.total == 1
    assert (
        abs(dataset_result.map_50_stats.mean - sample_result.map_50)
        < FLOATING_POINT_TOLERANCE
    )
    assert (
        abs(dataset_result.map_50_stats.median - sample_result.map_50)
        < FLOATING_POINT_TOLERANCE
    )
    assert dataset_result.map_50_stats.std == 0.0

    assert dataset_result.map_75_stats.total == 1
    assert (
        abs(dataset_result.map_75_stats.mean - sample_result.map_75)
        < FLOATING_POINT_TOLERANCE
    )
    assert (
        abs(dataset_result.map_75_stats.median - sample_result.map_75)
        < FLOATING_POINT_TOLERANCE
    )
    assert dataset_result.map_75_stats.std == 0.0


if __name__ == "__main__":
    test_map_layout_evaluations()



import json
from pathlib import Path

import numpy as np
from docling_metrics_layout.layout_types import (
    BboxResolution,
    LayoutMetricSample,
    MultiLabelMatrixEvaluation,
)
from docling_metrics_layout.tore.multi_label_confusion_matrix import (
    MultiLabelConfusionMatrix,
)

# Get the directory of this test file
TEST_DATA_DIR = Path(__file__).parent / "data"


def test_multi_label_confusion_matrix():
    # Test in microscopic images with artificial data
    canonical_categories = [0, 1, 2]
    categories_names = {
        0: "classA",
        1: "classB",
        2: "classC",
    }
    image_width = 10
    image_height = 12

    # BboxResolution expects (x1, y1, x2, y2) format
    gt_resolutions = [
        BboxResolution(category_id=0, bbox=[1, 1, 3.1, 3]),
        BboxResolution(category_id=1, bbox=[7, 7, 9, 9]),
        # GT that does not exist in predictions
        BboxResolution(category_id=2, bbox=[1, 7, 3, 9]),
    ]
    pred_resolutions = [
        # Exact prediction on bbox1
        BboxResolution(category_id=0, bbox=[1, 1, 3.1, 3]),
        # For bbox2 the correct class is found
        BboxResolution(category_id=1, bbox=[7, 7, 9, 9]),
        # Two extra wrong class is detected to overlap with bbox2 and lower
        BboxResolution(category_id=2, bbox=[7, 8, 9, 11]),
        BboxResolution(category_id=0, bbox=[7, 8, 9, 11]),
    ]

    mcm = MultiLabelConfusionMatrix()
    gt = mcm.make_binary_representation(image_width, image_height, gt_resolutions)
    pred = mcm.make_binary_representation(image_width, image_height, pred_resolutions)
    print(gt)
    print(pred)

    confusion_matrix = mcm.generate_confusion_matrix(gt, pred, canonical_categories)
    print("Confusion matrix:")
    print(confusion_matrix)

    metrics = mcm.compute_metrics(confusion_matrix, categories_names)
    print("Metrics:")
    print(metrics)


def test_multi_label_confusion_matrix_paper():
    r"""
    Testing the MultiLabelConfusionMatrix with the example from the paper
    https://csitcp.org/paper/10/108csit01.pdf
    """
    categories_names = {
        0: "classA",
        1: "classB",
        2: "classC",
        3: "classD",
    }
    categories_ids = list(categories_names.keys())
    image_width = 7
    image_height = 1

    # BboxResolution expects (x1, y1, x2, y2) format
    gt_resolutions = [
        # x1: 1 1 0 0
        BboxResolution(category_id=1, bbox=[0, 0, 1, 1]),
        BboxResolution(category_id=0, bbox=[0, 0, 1, 1]),
        # x2: 0 1 1 0
        BboxResolution(category_id=2, bbox=[1, 0, 2, 1]),
        BboxResolution(category_id=1, bbox=[1, 0, 2, 1]),
        # x3: 0 0 0 1
        BboxResolution(category_id=3, bbox=[2, 0, 3, 1]),
        # x4: 1 1 1 1
        BboxResolution(category_id=3, bbox=[3, 0, 4, 1]),
        BboxResolution(category_id=2, bbox=[3, 0, 4, 1]),
        BboxResolution(category_id=1, bbox=[3, 0, 4, 1]),
        BboxResolution(category_id=0, bbox=[3, 0, 4, 1]),
        # x5: 0 1 1 0
        BboxResolution(category_id=2, bbox=[4, 0, 5, 1]),
        BboxResolution(category_id=1, bbox=[4, 0, 5, 1]),
        # x6: 0 1 1 0
        BboxResolution(category_id=2, bbox=[5, 0, 6, 1]),
        BboxResolution(category_id=1, bbox=[5, 0, 6, 1]),
        # x7: 0 1 0 1
        BboxResolution(category_id=3, bbox=[6, 0, 7, 1]),
        BboxResolution(category_id=1, bbox=[6, 0, 7, 1]),
    ]
    pred_resolutions = [
        # x1: 1 1 0 0
        BboxResolution(category_id=1, bbox=[0, 0, 1, 1]),
        BboxResolution(category_id=0, bbox=[0, 0, 1, 1]),
        # x2: 1 1 1 0
        BboxResolution(category_id=2, bbox=[1, 0, 2, 1]),
        BboxResolution(category_id=1, bbox=[1, 0, 2, 1]),
        BboxResolution(category_id=0, bbox=[1, 0, 2, 1]),
        # x3: 1 0 0 1
        BboxResolution(category_id=3, bbox=[2, 0, 3, 1]),
        BboxResolution(category_id=0, bbox=[2, 0, 3, 1]),
        # x4: 0 1 1 1
        BboxResolution(category_id=3, bbox=[3, 0, 4, 1]),
        BboxResolution(category_id=2, bbox=[3, 0, 4, 1]),
        BboxResolution(category_id=1, bbox=[3, 0, 4, 1]),
        # x5: 0 0 1 0
        BboxResolution(category_id=1, bbox=[4, 0, 5, 1]),
        # x6: 1 1 0 0
        BboxResolution(category_id=1, bbox=[5, 0, 6, 1]),
        BboxResolution(category_id=0, bbox=[5, 0, 6, 1]),
        # x7: 1 0 1 0
        BboxResolution(category_id=2, bbox=[6, 0, 7, 1]),
        BboxResolution(category_id=0, bbox=[6, 0, 7, 1]),
    ]

    # Initialize MultiLabelConfusionMatrix
    mcm = MultiLabelConfusionMatrix()

    # Make the binary representations without adding the background class
    gt = mcm.make_binary_representation(
        image_width, image_height, gt_resolutions, set_background=False
    )
    pred = mcm.make_binary_representation(
        image_width, image_height, pred_resolutions, set_background=False
    )
    print(gt)
    print()
    print(pred)

    # Compute the confusion matrix
    confusion_matrix = mcm.generate_confusion_matrix(gt, pred, categories_ids)
    print("Confusion matrix:")
    print(confusion_matrix)
    print()

    # Compute metrics
    metrics: MultiLabelMatrixEvaluation = mcm.compute_metrics(
        confusion_matrix, categories_names
    )
    precision_matrix = metrics.detailed.precision_matrix
    recall_matrix = metrics.detailed.recall_matrix
    print("Precision matrix:")
    print(precision_matrix)
    print()
    print("Recall matrix:")
    print(recall_matrix)

    # Evaluate with the numbers from the paper
    correct_confusion_matrix = np.asarray(
        [
            [1.00, 0.33, 0.33, 0.33],
            [0.83, 4.67, 0.50, 0.00],
            [1.33, 1.00, 1.67, 0.00],
            [1.00, 0.00, 0.50, 1.50],
        ]
    )
    correct_precision_matrix = np.asarray(
        [
            [0.24, 0.06, 0.11, 0.18],
            [0.20, 0.78, 0.17, 0.00],
            [0.32, 0.17, 0.56, 0.00],
            [0.24, 0.00, 0.17, 0.82],
        ]
    )
    correct_recall_matrix = np.asarray(
        [
            [0.50, 0.17, 0.17, 0.17],
            [0.14, 0.78, 0.08, 0.00],
            [0.33, 0.25, 0.42, 0.00],
            [0.33, 0.00, 0.17, 0.50],
        ]
    )

    precision_col_sums = np.sum(precision_matrix, axis=0)
    recall_row_sums = np.sum(recall_matrix, axis=1)
    assert np.allclose(precision_col_sums, 1.0, rtol=0, atol=1e-08), (
        "Col sums of precision matrix must be 1"
    )
    assert np.allclose(recall_row_sums, 1.0, rtol=0, atol=1e-08), (
        "Row sums of recall matrix must be 1"
    )

    assert not np.allclose(
        confusion_matrix, correct_confusion_matrix, rtol=0, atol=1e-08
    ), "Wrong confusion matrix"
    assert not np.allclose(
        precision_matrix, correct_precision_matrix, rtol=0, atol=1e-08
    ), "Wrong precision matrix"
    assert not np.allclose(recall_matrix, correct_recall_matrix, rtol=0, atol=1e-08), (
        "Wrong recall matrix"
    )


def test_preds_on_preds():
    r"""
    Confusion matrix of preds vs preds (identical)

    Expected confusion matrix: Only the diagonal elements must be non-zero
    """
    # Load test layout sample
    test_data_fn = TEST_DATA_DIR / "dlnv1_t1_preds_score.json"
    with open(test_data_fn, "r") as fd:
        sample_dict = json.load(fd)
        sample = LayoutMetricSample.model_validate(sample_dict)

    # Initialize MultiLabelConfusionMatrix
    mcm = MultiLabelConfusionMatrix()

    # Make binary representations using page_resolution_a
    binary_preds = mcm.make_binary_representation(
        sample.page_width, sample.page_height, sample.page_resolution_a
    )

    # Compute the confusion matrix (preds vs preds)
    categories = list({res.category_id for res in sample.page_resolution_a})
    confusion_matrix = mcm.generate_confusion_matrix(
        binary_preds, binary_preds, sorted(categories)
    )

    # All non-diagonal values must be zero
    e = np.eye(len(categories))
    assert np.all(confusion_matrix * e == confusion_matrix)


def test_preds_on_empty():
    r"""
    Confusion matrix preds on gt without any predictions (not even background)

    Expected confusion matrix: All elements must be zero
    """
    # Load test layout sample
    test_data_fn = TEST_DATA_DIR / "dlnv1_t1_preds_score.json"
    with open(test_data_fn, "r") as fd:
        sample_dict = json.load(fd)
        sample = LayoutMetricSample.model_validate(sample_dict)

    # Initialize MultiLabelConfusionMatrix
    mcm = MultiLabelConfusionMatrix()

    # Make binary representations
    binary_gt = mcm.make_binary_representation(
        sample.page_width, sample.page_height, [], set_background=False
    )
    binary_preds = mcm.make_binary_representation(
        sample.page_width, sample.page_height, sample.page_resolution_a
    )

    # Compute the confusion matrix (empty gt vs preds)
    categories = list({res.category_id for res in sample.page_resolution_a})
    confusion_matrix = mcm.generate_confusion_matrix(
        binary_gt, binary_preds, sorted(categories)
    )

    # Confusion matrix must be all zeros
    zeros = np.zeros_like(confusion_matrix)
    assert np.all(confusion_matrix == zeros)


if __name__ == "__main__":
    test_multi_label_confusion_matrix()
    test_multi_label_confusion_matrix_paper()
    test_preds_on_preds()
    test_preds_on_empty()
